#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#if defined(HAVE_CLAMAV)
# include <clamav.h>
#else
# warning "No compile-time antivirus scanning method configured."
#endif
#include <glib.h>
#if defined(_WIN32)
# include <windows.h>
#else
# include <unistd.h>
#endif
#ifdef HAVE_LIBENDEAVOUR2
#include <endeavour2.h>
#endif

#include "../include/disk.h"

#include "avs_avengine_scan.h"
#include "config.h"


typedef struct _AVSSignalData		AVSSignalData;
#if defined(HAVE_CLAMAV)
typedef struct _AVSClamAVData		AVSClamAVData;
#endif

/* Output */
static void avs_print_error(const gchar *msg);
static void avs_print_scanning_file(
	const gchar *path,
	const gulong i, const gulong n
);
static void avs_print_scanned(
	const gulong files_scanned,
	const gulong data_scanned_blocks,
	const gulong total_data_scanned_blocks
);
static void avs_print_scanning_directory(const gchar *path);
static void avs_print_clean(const gchar *path);
static void avs_print_infected(const gchar *path, const gchar *virus);
static void avs_print_problem(const gchar *path, const gchar *problem);


/* Utilities */
static gboolean avs_is_executable_path_mode(
	const gchar *path,
	const mode_t m
);
static gchar *avs_strdup_shorten(
	const gchar *s,
	const gint max_characters
);


/* Callbacks */
static void avs_avengine_scan_signal_cb(int s);


#if defined(HAVE_CLAMAV)
/* ClamAV */
static AVSClamAVData *avs_clamav_init(
	const gchar *db_path,
	AVSAVEngineSettings *engine_settings,
	gint *status_rtn
);
static void avs_clamav_apply_engine_settings(
	AVSClamAVData *d,
	struct cl_engine *engine,
	AVSAVEngineSettings *engine_settings
);
static void avs_clamav_shutdown(AVSClamAVData *d);
static gint avs_clamav_scan_file(
	AVSClamAVData *d,
	const gchar *path,
	gint *stop_count,
	AVSAVEngineSettings *engine_settings
);
static gint avs_clamav_scan_directory(
	AVSClamAVData *d,
	const gchar *path,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gboolean *reload,
	gint *stop_count,
	gulong *directory_level,
	AVSAVEngineSettings *engine_settings
);
#endif	/* HAVE_CLAMAV */


/* Calculate Total */
static gulong avs_calculate_total_objects_iterate(
	const gchar *path,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gint *stop_count
);
static gulong avs_calculate_total_objects(
	GList *paths_list,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gint *stop_count
);

/* Get Antivirus Engine Version */
gchar *avs_get_avengine_name_version_string(void);

/* Antivirus Engine Scan */
gint avs_avengine_scan(
	const gchar *db_path,
	GList *paths_list,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gboolean *reload,
	gint *stop_count,
	AVSAVEngineSettings *engine_settings
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


#if defined(HAVE_CLAMAV)
/*
 *	ClamAV Data:
 */
struct _AVSClamAVData {
	gchar		*db_path;		/* Database file or directory */
 	gchar		*db_directory_path;	/* Database directory */
	struct cl_engine	*engine;	/* Database handle */
	gulong		ndb_entries;		/* Entries in the database */
	struct cl_stat	db_stat;		/* Database stats, for
						 * checking database changes */
	gulong		files_scanned,		/* Files scanned so far */
			size_scanned_blocks,
			infected_files,
			total_files;		/* Total files to be scanned */
};
#define AVS_CLAMAV_DATA(p)		((AVSClamAVData *)(p))
#endif	/* HAVE_CLAMAV */

/*
 *	UNIX Signal Callback Data:
 */
struct _AVSSignalData {
	gint		stop_count;
	gboolean	reload;
};
#define AVS_SIGNAL_DATA(p)		((AVSSignalData *)(p))
static AVSSignalData	avs_signal_data;


/*
 *	Prints messages to the calling process.
 *
 *	This module is intended to be called from a separate process,
 *	it needs to print messages to the calling process in order to
 *	report its status.
 *
 *	Both stdout (standard output) and stderr (standard error)
 *	messages are printed and read by the calling process.
 *
 *	All stdout messages must be printed and parsed in the format:
 *
 *	"<prefix> <value>\n"
 *
 *	Where <prefix> determines the message type and how to parse
 *	the <value>. All stdout messages end with a newline character.
 *
 *	All stderr messages must be printed and parsed in the format:
 *
 *	"<value>\n"
 *
 *	They do not have any prefix that determines any type and
 *	therefore should be parsed and displayed as just one simple
 *	string. All stderr messages end with a newline character.
 */
static void avs_print_scanning_file(
	const gchar *path,
	const gulong i, const gulong n
)
{
	g_print(
		"Scanning file: \"%s\" %ld %ld\n",
		path,
		i,
		n
	);
}

static void avs_print_scanned(
	const gulong files_scanned,
	const gulong data_scanned_blocks,
	const gulong total_data_scanned_blocks
)
{
	g_print(
		"Scanned: %ld %ld %ld\n",
		files_scanned,
		data_scanned_blocks,
		total_data_scanned_blocks
	);
}

static void avs_print_scanning_directory(const gchar *path)
{
	g_print(
		"Scanning directory: \"%s\"\n",
		path
	);
}

static void avs_print_clean(const gchar *path)
{
	g_print(
		"Clean: \"%s\"\n",
		path
	);
}

static void avs_print_infected(const gchar *path, const gchar *virus)
{
	g_print(
		"Infected: \"%s\" %s\n",
		path,
		virus
	);
}

static void avs_print_problem(const gchar *path, const gchar *problem)
{
	g_print(
		"Problem: \"%s\" %s\n",
		path,
		problem
	);
}

static void avs_print_error(const gchar *msg)
{
	g_printerr(
		"%s\n",
		msg
	);
}


/*
 *	Checks if the object is an executable.
 */
static gboolean avs_is_executable_path_mode(
	const gchar *path,
	const mode_t m
)
{
#if defined(_WIN32)
	const gchar	*ext,
			*name;

	if(path == NULL)
		return(FALSE);

	/* Get the name */
	name = g_basename(path);
	if(name == NULL)
		name = path;

	/* Get the extension */
	ext = (const gchar *)strchr(
		(const char *)name,
		'.'
	);

	/* No extension? */
	if(ext == NULL)
		return(FALSE);

	/* Seek past the '.' deliminator */
	ext++;

	/* The following Win32 extensions are considered "executable" */
	if(!g_strcasecmp(ext, "bat") ||
	   !g_strcasecmp(ext, "com") ||
	   !g_strcasecmp(ext, "exe") ||
	   !g_strcasecmp(ext, "bin") ||
	   !g_strcasecmp(ext, "dll") ||
	   !g_strcasecmp(ext, "lib") ||
	   !g_strcasecmp(ext, "so") ||
	   !g_strcasecmp(ext, "o") ||
	   !g_strcasecmp(ext, "a")
	)
		return(TRUE);
	else
		return(FALSE);
#else
	if((m & S_IXUSR) || (m & S_IXGRP) || (m & S_IXOTH))
		return(TRUE);
	else
		return(FALSE);
#endif
}


/*
 *	UNIX signal callback.
 */
static void avs_avengine_scan_signal_cb(int s)
{
	AVSSignalData *d = &avs_signal_data;
	switch(s)
	{
#ifdef SIGHUP
	  case SIGHUP:
		if(d != NULL)
			d->reload = TRUE;
		break;
#endif

#ifdef SIGINT
	  case SIGINT:
		if(d != NULL)
		{
			d->stop_count++;
		}
		else
		{
			exit(4);		/* User aborted */
		}
		break;
#endif

#ifdef SIGTERM
	  case SIGTERM:
		exit(4);			/* User aborted */
		break;
#endif

#ifdef SIGSEGV
	  case SIGSEGV:
		exit(8);			/* Segmentation fault */
		break;
#endif

#ifdef SIGPIPE
	  case SIGPIPE:
		break;
#endif
	}
}


/*
 *	Coppies the string any shortens it if it exceeds the maximum
 *	length.
 *
 *	The s specifies the string to copy.
 *
 *	The max_characters specifies the maximum length of characters.
 *
 *	Returns a copy of s that may have been shortened.
 */
static gchar *avs_strdup_shorten(
	const gchar *s,
	const gint max_characters
)
{
	gint len;

	if(s == NULL)
		return(NULL);

	len = (gint)strlen((const char *)s);
	if((len > max_characters) && (max_characters > 3))
	{
		/* Need to shorten the string */
		const gint i = len - max_characters + 3;
		return(g_strconcat(
			"...",
			(const gchar *)(s + i),
			NULL
		));
	}      
	else
	{   
		return(g_strdup(s));
	}
}

/*
 *	Joins two paths.
 */
static gchar *avs_join_paths(
	const gchar *parent_path,
	const gchar *child_path
)
{
#ifdef HAVE_LIBENDEAVOUR2
	return(edv_paths_join(
		parent_path,
		child_path
	));
#else
	if(g_path_is_absolute(child_path))
		return(g_strdup(child_path));
	else
		return(g_strconcat(
			parent_path,
			G_DIR_SEPARATOR_S,
			child_path,
			NULL
		));
#endif
}

/*
 *	Gets the link's target value.
 */
static gchar *avs_get_link_target(const gchar *path)
{
#ifdef HAVE_LIBENDEAVOUR2
	return(edv_link_get_target(path));
#else
	return((gchar *)GetAllocLinkDest((const char *)path));
#endif
}


#if defined(HAVE_CLAMAV)
/*
 *	Allocates, loads, and builds a new cl_engine (ClamAV Virus
 *	Database) and creates a new AVSClamAVData to store it on.
 *
 *	This function does not call cl_init(), cl_init() should be
 *	called once by the calling function.
 *
 *	The db_path specifies the path to the ClamAV Virus Database
 *	file or directory that will be loaded and built.
 *
 *	The engine_settings specifies the AVSAVEngineSettings that
 *	will be applied to the loaded and built cl_engine.
 *
 *	If status_rtn is not NULL then *status_rtn will be set to 0
 *	on success or non-zero on error.
 *
 *	Returns the new AVSClamAVData or NULL on error.
 */
static AVSClamAVData *avs_clamav_init(
	const gchar *db_path,
	AVSAVEngineSettings *engine_settings,
	gint *status_rtn
)
{
	unsigned int u32;
	struct stat stat_buf;
	gint status;
	gchar *s;
	AVSClamAVData *d = AVS_CLAMAV_DATA(g_malloc0(sizeof(AVSClamAVData)));
	if(d == NULL)
	{
		avs_print_error("Memory allocation error");
		if(status_rtn != NULL)
			*status_rtn = -3;
		return(NULL);
	}

	/* If no virus database was specified then get the default
	 * (hard coded) virus database set by ClamAV
	 */
	if(db_path == NULL)
		db_path = cl_retdbdir();
	if(db_path == NULL)
	{
		avs_print_error("Virus database location was not specified");
		g_free(d);
		if(status_rtn != NULL)
			*status_rtn = -50;
		return(NULL);
	}
	d->db_path = STRDUP(db_path);
	if(d->db_path == NULL)
	{
		avs_print_error("Memory allocation error");
		g_free(d);
		if(status_rtn != NULL)
			*status_rtn = -3;
		return(NULL);
	}

	/* Get the virus database's stats */
	if(stat((const char *)d->db_path, &stat_buf))
	{
		const gint error_code = (gint)errno;
		avs_print_problem(d->db_path, g_strerror(error_code));
		g_free(d->db_path);
		g_free(d);
		if(status_rtn != NULL)
			*status_rtn = -50;
		return(NULL);
	}

	s = avs_strdup_shorten(d->db_path, 50);
	g_print("Loading virus database \"%s\"...\n", s);
	g_free(s);

	/* Create a new virus database */
	d->engine = cl_engine_new();
	if(d->engine == NULL)
	{
		avs_print_problem(d->db_path, "Unable to allocate memory to load the virus database");
		g_free(d->db_path);
		g_free(d);
		if(status_rtn != NULL)
			*status_rtn = -3;
		return(NULL);
	}

	/* Load the virus database */
	status = (gint)cl_load(
		d->db_path,
		d->engine,
		&u32,
		CL_DB_STDOPT
	);
	if(status != CL_SUCCESS)
	{
		gchar *s = g_strconcat(
			"Virus database load error: ",
			cl_strerror(status),
			NULL
		);
		avs_print_problem(d->db_path, s);
		g_free(s);
		(void)cl_engine_free(d->engine);
		g_free(d->db_path);
		g_free(d);
		if(status_rtn != NULL)
			*status_rtn = -50;
		return(NULL);
	}

	d->ndb_entries = (gulong)u32;

	/* Record the virus database direcotry path */
#ifdef S_ISDIR
	if(S_ISDIR(stat_buf.st_mode))
#else
	if(FALSE)
#endif
	{
		d->db_directory_path = STRDUP(d->db_path);
		if(d->db_directory_path == NULL)
		{
			avs_print_error("Memory allocation error");
			(void)cl_engine_free(d->engine);
			g_free(d->db_path);
			g_free(d);
			if(status_rtn != NULL)
				*status_rtn = -3;
			return(NULL);
		}
	}
	else
	{
		d->db_directory_path = g_dirname(d->db_path);
		if(d->db_directory_path == NULL)
		{
			avs_print_error("Memory allocation error");
			(void)cl_engine_free(d->engine);
			g_free(d->db_path);
			g_free(d);
			if(status_rtn != NULL)
				*status_rtn = -3;
			return(NULL);
		}
	}

	/* Build the virus database */
	status = (gint)cl_engine_compile(d->engine);
	if(status != CL_SUCCESS)
	{
		gchar *s = g_strconcat(
			"Virus database build error: ",
			cl_strerror(status),
			NULL
		);
		avs_print_problem(d->db_path, s);
		g_free(s);
		(void)cl_engine_free(d->engine);
		g_free(d->db_path);
		g_free(d->db_directory_path);
		g_free(d);
		if(status_rtn != NULL)
			*status_rtn = -50;
		return(NULL);
	}

	/* Apply the AVSAVEngineSettings to ClamAV */
	avs_clamav_apply_engine_settings(
		d,
		d->engine,
		engine_settings
	);

	/* Get the virus database directory statistics so that we can
	 * check if it has been modified and reload the virus database
	 */
	(void)memset(
		&d->db_stat,
		0x00,
		sizeof(d->db_stat)
	);
	if(cl_statinidir(
		d->db_directory_path,
		&d->db_stat
	))
	{
		avs_print_problem(
			d->db_directory_path,
			"Unable to get the virus database directory statistics"
		);
		(void)cl_engine_free(d->engine);
		g_free(d->db_path);
		g_free(d->db_directory_path);
		g_free(d);
		if(status_rtn != NULL)
			*status_rtn = -50;
		return(NULL);
	}

	if(status_rtn != NULL)
		*status_rtn = 0;

	return(d);
}

/*
 *	Applies the AVSAVEngineSettings to a newly loaded and built
 *	cl_engine (ClamAV Virus Database).
 *
 *	This should be called after a new cl_engine has been allocated,
 *	loaded, and built.
 *
 *	The d specifies the AVSClamAVData.
 *
 *	The engine specifies a newly built cl_engine.
 *
 *	The engine_settings specifies the AVSAVEngineSettings to be
 *	applied to the cl_engine.
 */
static void avs_clamav_apply_engine_settings(
	AVSClamAVData *d,
	struct cl_engine *engine,
	AVSAVEngineSettings *engine_settings
)
{
	/* Maximum Scan Size */
	if(engine_settings->max_scan_size != 0l)
		cl_engine_set_num(
			engine,
			CL_ENGINE_MAX_SCANSIZE,
			(long long)engine_settings->max_scan_size
		);

	/* Maximum Archive Files */
	if(engine_settings->max_archive_files != 0l)
		cl_engine_set_num(
			engine,
			CL_ENGINE_MAX_FILES,
			(long long)engine_settings->max_archive_files
		);

	/* Maximum Archive File Size */
	if(engine_settings->max_archive_file_size != 0l)
		cl_engine_set_num(
			engine,
			CL_ENGINE_MAX_FILESIZE,
			(long long)engine_settings->max_archive_file_size
		);

	/* Maximum Archive Recursions */
	if(engine_settings->max_archive_recursions != 0l)
		cl_engine_set_num(
			engine,
			CL_ENGINE_MAX_RECURSION,
			(long long)engine_settings->max_archive_recursions
		);
}

/*
 *	Shuts down the Clam AntiVirus.
 */
static void avs_clamav_shutdown(AVSClamAVData *d)
{
	if(d == NULL)
		return;

	/* Virus database */
	if(d->engine != NULL)
		(void)cl_engine_free(d->engine);

	/* Virus database stat */
	(void)cl_statfree(&d->db_stat);

	/* Shutdown ClamAV */
/*	cl_shutdown(); */

	g_free(d->db_path);
	g_free(d->db_directory_path);

	g_free(d);
}

/*
 *	Use the Clam AntiVirus to scan the specified file.
 */
static gint avs_clamav_scan_file(
	AVSClamAVData *d,
	const gchar *path,
	gint *stop_count,
	AVSAVEngineSettings *engine_settings
)
{
	unsigned long int scanned_blocks = 0l;
	gint		status,
			fd;
	guint options;
	gchar *s;
	const gchar *virus_name = NULL;
#ifdef CL_SCAN_GENERAL_HEURISTICS
	struct cl_scan_options scan_options = {CL_SCAN_GENERAL_HEURISTICS,
		                               ~0,
					       0,
					       0,
					       0};
#endif

	/* Interrupted? */
	if(*stop_count > 0)
		return(-4);

	avs_print_scanning_file(
		path,
		d->files_scanned,
		d->total_files
	);

	/* Open the file for reading */
	fd = open(
		(const char *)path,
		O_RDONLY
	);
	if(fd < 0)
	{
		/* Unable to open the file for reading, report error */
		const gint error_code = (gint)errno;
		avs_print_problem(path, g_strerror(error_code));

		/* Unable to open an input file */
		return(-54);
	}

	/* Interrupted? */
	if(*stop_count > 0)
	{
		/* Close the file but discard the return value because we
		 * are interrupting and thus do not care of buffered
		 * read()s or write()s
		 */
		(void)close(fd);
		return(-4);
	}

#if 0
/* ClamAV no longer uses limits */
	/* Set up archive limits */
	memset(&limits, 0x00, sizeof(struct cl_limits));

	/* Maximum scan size for archives */
	limits.maxscansize = (unsigned long int)engine_settings->max_archive_size_bytes;

	/* Maximum scan size for compressed files */
	limits.maxfilesize = (unsigned long int)engine_settings->max_compressed_file_size_bytes;

	/* Maximum recursion level */
	limits.maxreclevel = (unsigned int)engine_settings->max_recursions;

	/* Maximum files to scan within an archive */
	limits.maxfiles = (unsigned int)engine_settings->max_files;

	/* Maximum mail recursion level */
	limits.maxmailrec = (unsigned int)engine_settings->max_mail_recursions;

	/* Maximum compression ratio */
	limits.maxratio = (unsigned int)engine_settings->max_compression_ratio;

	/* Disable memory limit for bzip2 and other unpackers */
	limits.archivememlim = 0;
#endif


	/* Set up scan options */
#if defined(CL_SCAN_STDOPT)
	options = CL_SCAN_STDOPT;
#elif defined(CL_SCAN_RAW) && defined(CL_SCAN_ARCHIVE) && defined(CL_SCAN_MAIL) && defined(CL_SCAN_HTML)
	options = CL_SCAN_RAW | CL_SCAN_ARCHIVE | CL_SCAN_MAIL |
			  CL_SCAN_HTML;
#elif defined(CL_SCAN_ARCHIVE) && defined(CL_SCAN_MAIL)
	options = CL_SCAN_ARCHIVE | CL_SCAN_MAIL;
#elif defined(CL_SCAN_RAW)
	options = CL_SCAN_RAW;
#else
#warning No supported CL_SCAN_* options are #defined, options will always be 0.
 	options = 0;
#endif

	/* Scan this file */
	status = (gint)cl_scandesc(
		fd,
#ifdef CL_SCAN_GENERAL_HEURISTICS
		path,
#endif
		&virus_name,
		&scanned_blocks,
		d->engine,
#ifdef CL_SCAN_GENERAL_HEURISTICS
		&scan_options
#else
		(unsigned int)options
#endif
	);

	/* Close the file */
	if(close(fd))
	{
		/* Unable to close the file, report problem */
		const gint error_code = (gint)errno;
		avs_print_problem(path, g_strerror(error_code));

		/* An error occured while reading an input file */
		return(-55);
	}

	/* Count this file as scanned and how much scanned */
	d->files_scanned++;
	d->size_scanned_blocks += (gulong)scanned_blocks;
	avs_print_scanned(
		d->files_scanned,
		(gulong)(scanned_blocks *
			((gfloat)CL_COUNT_PRECISION / 1024.0f)
		),
		(gulong)((gfloat)d->size_scanned_blocks *
			((gfloat)CL_COUNT_PRECISION / 1024.0f)
		)
	);

	/* Check the scan results */
	switch((cl_error_t)status)
	{
#if CL_CLEAN != CL_SUCCESS
	    case CL_CLEAN:
#endif
	    case CL_SUCCESS:
		/* Interrupted? */
		if(*stop_count > 0)
			return(-4);
		/* Print the clean message only if reporting all objects */
		if(engine_settings->flags & AVS_AVENGINE_REPORT_ALL_OBJECTS)
			avs_print_clean(path);
		status = 0;
		break;

	    case CL_VIRUS:
		/* Detected a virus, print the virus message */
		avs_print_infected(path, virus_name);
		/* Count this infected file */
		d->infected_files++;
		status = -1;
		break;

	    case CL_ENULLARG:
	    case CL_EARG:
	    case CL_EMALFDB:
	    case CL_ECVD:
	    case CL_EVERIFY:
	    case CL_EUNPACK:
		s = STRDUP(cl_strerror((int)status));
		if(s != NULL)
		{
			avs_print_problem(path, s);
			g_free(s);
		}
		status = -2;
		break;

	    case CL_EOPEN:
	    case CL_ECREAT:
	    case CL_EUNLINK:
	    case CL_ESTAT:
	    case CL_EREAD:
	    case CL_ESEEK:
	    case CL_EWRITE:
	    case CL_EDUP:
	    case CL_EACCES:
	    case CL_ETMPFILE:
	    case CL_ETMPDIR:
	    case CL_EMAP:
	    case CL_EMEM:
	    case CL_ETIMEOUT:
		s = STRDUP(cl_strerror((int)status));
		if(s != NULL)
		{
			avs_print_problem(path, s);
			g_free(s);
		}
		status = -3;
		break;

	    case CL_BREAK:
		status = -4;
		break;

	    case CL_EMAXREC:
	    case CL_EMAXSIZE:
	    case CL_EMAXFILES:
	    case CL_EFORMAT:
		s = STRDUP(cl_strerror((int)status));
		if(s != NULL)
		{
			avs_print_problem(path, s);
			g_free(s);
		}
		status = -57;			/* Scan of object was
						 * incomplete because it would
						 * otherwise exceed a specified
						 * limit */
		break;
	}

	return(status);
}

/*
 *	Use Clam AntiVirus to scan the specified directory.
 */
static gint avs_clamav_scan_directory(
	AVSClamAVData *d,
	const gchar *path,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gboolean *reload,
	gint *stop_count,
	gulong *directory_level,
	AVSAVEngineSettings *engine_settings
)
{
	mode_t m;
	struct dirent *dent;
	struct stat stat_buf;
	DIR *dp;
	gint		status,
			status2;
	const gchar *s;
	gchar *child;

	/* Interrupted? */
	if(*stop_count > 0)
		return(-4);

	/* Directory recursion level reached? */
	if(engine_settings->max_directory_recursions > 0l)
	{
		if(*directory_level >= engine_settings->max_directory_recursions)
		{
			gchar *s = g_strdup_printf(
"Skipped, directory recursion maximum %ld reached",
				engine_settings->max_directory_recursions
			);
			avs_print_problem(path, s);
			g_free(s);
			return(-57);
		}
	}

	/* Open the directory */
	dp = opendir(path);
	if(dp == NULL)
	{
		const gint error_code = (gint)errno;
		avs_print_problem(path, g_strerror(error_code));
		return(-53);
	}

	/* Interrupted? */
	if(*stop_count > 0)
	{
		(void)closedir(dp);
		return(-4);
	}

	/* Scan directory */
	avs_print_scanning_directory(path);
	status = 0;
	while(*stop_count == 0)
	{
		dent = readdir(dp);
		if(dent == NULL)
			break;

		s = (const gchar *)dent->d_name;

		/* Check for virus database change or a request to reload */
		if(*reload || (cl_statchkdir(&d->db_stat) == 1))
		{
			/* Reload the virus database */
			unsigned int u32;
			struct stat stat_buf;
			gchar *s = avs_strdup_shorten(d->db_path, 50);
			g_print("Reloading virus database \"%s\"...\n", s);
			g_free(s);

			/* Reset the reload marker */
			if(*reload)
				*reload = FALSE;

			/* Delete the existing virus database */
			if(d->engine != NULL)
			{
				(void)cl_engine_free(d->engine);
				d->engine = NULL;
				d->ndb_entries = 0l;
			}

			/* Interrupted? */
			if(*stop_count > 0)
				break;

			if(stat((const char *)d->db_path, &stat_buf))
			{
				const gint error_code = (gint)errno;
				avs_print_problem(d->db_path, g_strerror(error_code));
				status = -50;
				break;
			}

			/* Rebuild the virus database */
			d->engine = cl_engine_new();
			if(d->engine == NULL)
			{
				avs_print_problem(d->db_path, "Unable to allocate memory for the virus database");
				status = -3;
				break;
			}
			status2 = cl_load(
				d->db_path,
				d->engine,
				&u32,
				CL_DB_STDOPT
			);
			if(status2 != CL_SUCCESS)
			{
				gchar *s = g_strconcat(
					"Virus database reload error: ",
					cl_strerror(status2),
					NULL
				);
				avs_print_problem(d->db_path, s);
				g_free(s);
				(void)cl_engine_free(d->engine);
				d->engine = NULL;
				d->ndb_entries = 0l;
				status = -50;
				break;
			}
			d->ndb_entries = (gulong)u32;
			status2 = cl_engine_compile(d->engine);
			if(status2 != CL_SUCCESS)
			{
				gchar *s = g_strconcat(
					"Virus database reload error: ",
					cl_strerror(status2),
					NULL
				);
				avs_print_problem(d->db_path, s);
				g_free(s);
				(void)cl_engine_free(d->engine);
				d->engine = NULL;
				d->ndb_entries = 0l;
				status = -50;
				break;
			}

		        /* Reapply the AVSAVEngineSettings to ClamAV */
		        avs_clamav_apply_engine_settings(
		                d,
		                d->engine,
		                engine_settings
		        );

			/* Must reget the virus database statistics
			 * after reloading it
			 */
			(void)cl_statfree(&d->db_stat);
			(void)memset(
				&d->db_stat,
				0x00,
				sizeof(d->db_stat)
			);
			if(cl_statinidir(
				d->db_directory_path,
				&d->db_stat
			))
			{
				avs_print_problem(
					d->db_directory_path,
					"Unable to get the virus database directory statistics"
				);
				(void)cl_engine_free(d->engine);
				d->engine = NULL;
				d->ndb_entries = 0l;
				status = -50;
				break;
			}

			/* Interrupted? */
			if(*stop_count > 0)
				break;
		}

		/* Skip the current and parent directory notations */
		if(!strcmp((const char *)s, ".") ||
		   !strcmp((const char *)s, "..")
		)
			continue;

		/* Get the full path to the child object */
		child = avs_join_paths(
			path,
			s
		);
		if(child == NULL)
		{
			if(status == 0)
				status = -3;
			continue;
		}

		/* Get this object's local stats */
		if(lstat((const char *)child, &stat_buf))
		{
			const gint error_code = (gint)errno;
			avs_print_problem(child, g_strerror(error_code));
			g_free(child);
			if(status == 0)
				status = -56;
			continue;
		}

		m = stat_buf.st_mode;

		/* Link? */
#ifdef S_ISLNK
		if(S_ISLNK(m))
#else
		if(FALSE)
#endif
		{
			gchar *target;

			/* Ignore links? */
			if(ignore_links)
			{
				g_free(child);
				continue;
			}

			/* Get the destination value and check if it is a
			 * reference to the current or parent directory
			 */
			target = avs_get_link_target(child);
			if(target != NULL)
			{
				if(!strcmp((const char *)target, ".") ||
				   !strcmp((const char *)target, "..")
				)
				{
					/* Skip this to prevent infinite recursion */
					g_free(child);
					g_free(target);
					continue;
				}
				g_free(target);
			}
		}

		/* Get this object's destination stats */
		if(stat((const char *)child, &stat_buf))
		{
			const gint error_code = (gint)errno;
			avs_print_problem(child, g_strerror(error_code));
			g_free(child);
			if(status == 0)
				status = -56;
			continue;
		}

		m = stat_buf.st_mode;

		/* Directory */
#ifdef S_ISDIR
		if(S_ISDIR(m))
#else
		if(FALSE)
#endif
		{
			if(recursive)
			{
				*directory_level = *directory_level + 1l;
				status2 = avs_clamav_scan_directory(
					d,
					child,
					recursive,
					executables_only,
					ignore_links,
					reload,
					stop_count,
					directory_level,
					engine_settings
				);
				*directory_level = *directory_level - 1l;
				if((status2 != 0) && (status == 0))
					status = status2;
			}
		}
		/* Regular File */
#ifdef S_ISREG
		else if(S_ISREG(m))
#else
		else if(TRUE)
#endif
		{
			if(executables_only ?
				avs_is_executable_path_mode(child, m) : TRUE
			)
			{
				status2 = avs_clamav_scan_file(
					d,
					child,
					stop_count,
					engine_settings
				);
				if((status2 != 0) && (status == 0))
					status = status2;
			}
		}

		g_free(child);
	}

	/* Close the directory */
	(void)closedir(dp);

	/* User aborted? */
	if(*stop_count > 0)
	{
		if(status == 0)
			status = -4;
	}

	return(status);
}
#endif	/* HAVE_CLAMAV */


/*
 *	Calculates the total files to be scanned in the directory
 *	specified by path.
 *
 *	If recursive is TRUE then files to be scanned in subdirectories
 *	of the specified directory will be counted.
 *
 *	If executables_only is TRUE then only files that are set
 *	executable will be counted.
 */
static gulong avs_calculate_total_objects_iterate(
	const gchar *path,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gint *stop_count
)
{
	mode_t m;
	gulong subtotal = 0l;
	gchar *child;
	const gchar *s;
	struct dirent *dent;
	struct stat stat_buf;
	DIR *dp = opendir(path);
	if(dp == NULL)
		return(subtotal);

	while(*stop_count == 0)
	{
		dent = readdir(dp);
		if(dent == NULL)
			break;

		s = (const gchar *)dent->d_name;

		/* Skip the current and parent directory notations */
		if(!strcmp((const char *)s, ".") ||
		   !strcmp((const char *)s, "..")
		)
			continue;

		/* Get the full path to the child object */
		child = avs_join_paths(
			path,
			s
		);
		if(child == NULL)
			continue;

		/* Get the object's local stats */
		if(lstat((const char *)child, &stat_buf))
		{
			/* Unable to get the local stats, do not count this */
			g_free(child);
			continue;
		}

		m = stat_buf.st_mode;

#ifdef S_ISLNK
		/* Link? */
		if(S_ISLNK(m))
#else
		if(FALSE)
#endif
		{
			gchar *dest;

			/* Ignore links? */
			if(ignore_links)
			{
				/* Do not count this */
				g_free(child);
				continue;
			}

			/* Get the destination value and check if it is a
			 * reference to the current or parent directory
			 */
			dest = (gchar *)GetAllocLinkDest((const char *)child);
			if(dest != NULL)
			{
				if(!strcmp((const char *)dest, ".") ||
				   !strcmp((const char *)dest, "..")
				)
				{
					/* Skip this to prevent infinite recursion */
					g_free(child);
					g_free(dest);
					continue;
				}
				g_free(dest);
			}
		}

		/* Get this object's destination stats */
		if(stat((const char *)child, &stat_buf))
		{
			/* Unable to get the destination stats, do not count this */
			g_free(child);
			continue;
		}

		m = stat_buf.st_mode;

		/* Directory */
#ifdef S_ISDIR
		if(S_ISDIR(m))
#else
		if(FALSE)
#endif
		{
			if(recursive)
				subtotal += avs_calculate_total_objects_iterate(
					child,
					recursive,
					executables_only,
					ignore_links,
					stop_count
				);
		}
		/* Regular File */
#ifdef S_ISREG
		else if(S_ISREG(m))
#else
		else if(TRUE)
#endif
		{
			if(executables_only ?
				avs_is_executable_path_mode(child, m) : TRUE
			)
				subtotal++;
		}

		g_free(child);
	} 

	(void)closedir(dp);

	return(subtotal);
}

/*
 *	Calculates the total number of files to be scanned from the
 *	specified list of paths.
 *
 *	If a file is not accessable to be scanned in the specified
 *	path then it will be excluded from the returned total.
 */
static gulong avs_calculate_total_objects(    
	GList *paths_list,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gint *stop_count
)
{
	mode_t m;
	struct stat stat_buf;
	gulong total = 0l;
	const gchar *path;
	GList *glist;

	if(paths_list == NULL)
		return(total);

	/* Iterate through each object in the paths list */
	for(glist = paths_list; glist != NULL; glist = g_list_next(glist))
	{
		/* Interrupted */
		if(*stop_count > 0)
			break;

		path = (const gchar *)glist->data;
		if(STRISEMPTY(path))
			continue;

		/* Get this object's local stats */
		if(lstat((const char *)path, &stat_buf))
		{
			/* Unable to get local stats, do not count this */
			continue;
		}

		m = stat_buf.st_mode;

		/* Link */
#ifdef S_ISLNK
		if(S_ISLNK(m))
#else
		if(FALSE)
#endif
		{
			/* Links ignored? */
			if(ignore_links)
			{
				/* Do not count this */
				continue;
			}

			/* Get this link's destination stats */
			if(stat((const char *)path, &stat_buf))
			{
				/* Unable to get destination stats, do not count this */
				continue;
			}

			m = stat_buf.st_mode;

			/* Is this link's destination a directory? */
#ifdef S_ISDIR
			if(S_ISDIR(m))
#else
			if(FALSE)
#endif
			{
				total += avs_calculate_total_objects_iterate(
					path,
					recursive, executables_only, ignore_links,
					stop_count
				);
			}
			else
			{
				total++;
			}
		}
		/* Directory */
#ifdef S_ISDIR
		else if(S_ISDIR(m))
#else
		else if(FALSE)
#endif
		{
			total += avs_calculate_total_objects_iterate(
				path,
				recursive,
				executables_only,
				ignore_links,
				stop_count
			);
		}
		/* Regular File */
#ifdef S_ISREG
		else if(S_ISREG(m))
#else
		else if(TRUE)
#endif
		{
			total++;
		}
	}

	return(total);
}


/*
 *	Returns a dynamically allocated string describing the name and
 *	the version of the Antivirus Engine.
 */
gchar *avs_get_avengine_name_version_string(void)
{
#if defined(HAVE_CLAMAV)
	return(g_strdup_printf(
		"%s Version %s",
		"Clam AntiVirus",
		cl_retver()
	));
#else
	return(NULL);
#endif
}


/*
 *	Executes the AntiVirus Engine to perform a scan.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	General error or virus found.
 *	-2	Invalid value.
 *	-3	Systems error or memory allocation error.
 *	-4	User aborted (otherwise no other error).
 *	-5	User responded with "no" to a query (otherwise no
 *		other error).
 *	-6	An identical operation is already in progress.
 *	-7	Reserved.
 *	-8	Segmentation fault.
 *
 *	-50	Unable to load or initialize the virus database.
 *	-51	An input file has a corrupt format or a format not
 *		consistent with its extension.
 *	-52	Unsupported object type.
 *	-53	Unable to open an input directory.
 *	-54	Unable to open an input file.
 *	-55	An error occured while reading an input file.
 *	-56	Unable to stat() or lstat() an input file.
 *	-57	Scan of object is incomplete because it would
 *		otherwise exceed a specified limit.
 *	-58	Reserved.
 *	-59	Scanner subsystem or plugin unable to scan an input
 *		file.
 *
 *	-60	Unsupported file format.
 *	-61	Unable to lock the virus database directory.
 *
 *	-99	No virus scanner support compiled.
 */
gint avs_avengine_scan(
	const gchar *db_path,
	GList *paths_list,
	const gboolean recursive,
	const gboolean executables_only,
	const gboolean ignore_links,
	gboolean *reload,
	gint *stop_count,
	AVSAVEngineSettings *engine_settings
)
{
	gint status;
#if defined(HAVE_CLAMAV)
	AVSClamAVData *d;
#endif

	if((paths_list == NULL) || (engine_settings == NULL))
		return(-2);

	/* Interrupted? */
	if(*stop_count > 0)
		return(-4);

	/* Set the priority */
	if(TRUE)
	{
		/* Calculate the priority, negative values such as -20
		 * take up the most priority (all the cpu), while 20 takes
		 * up the least priority (least cpu)
		 */
#ifdef PRIO_MIN
		const int prio_highest = PRIO_MIN;
#else
#warning PRIO_MIN not defined, assuming PRIO_MIN is -20
		const int prio_highest = -20;
#endif
#ifdef PRIO_MAX
		const int prio_lowest = PRIO_MAX;
#else
#warning PRIO_MAX not defined, assuming PRIO_MAX is 20
		const int prio_lowest = 20;
#endif
		int prio = 0;
		const gchar *priority_name = "";
		switch(engine_settings->cpu_priority)
		{
		  case AVS_AVENGINE_CPU_PRIORITY_HIGHEST:
			priority_name = "Highest";
			prio = prio_highest;
			break;
		  case AVS_AVENGINE_CPU_PRIORITY_HIGH:
			priority_name = "High";
			prio = prio_highest / 2;
			break;
		  case AVS_AVENGINE_CPU_PRIORITY_MEDIUM:
			/* Default, leave prio as is */
			priority_name = "Medium";
			break;
		  case AVS_AVENGINE_CPU_PRIORITY_LOW:
			priority_name = "Low";
			prio = prio_lowest / 2;
			break;
		  case AVS_AVENGINE_CPU_PRIORITY_LOWEST:
			priority_name = "Lowest";
			prio = prio_lowest;
			break;
		}
		errno = 0;			/* Needed since setpriority()
						 * may return -1 on success */
		if(setpriority(
			PRIO_PROCESS,
			0,			/* This process */
			prio
		))
		{
			const gint error_code = (gint)errno;
			if(error_code != 0)
			{
				gchar *msg = g_strdup_printf(
"Unable to set CPU Priority to \"%s\" (%i): %s",
					priority_name,
					prio,
					g_strerror(error_code)
				);
				avs_print_error(msg);
				g_free(msg);
			}
		}
	}

	status = 0;

#if defined(HAVE_CLAMAV)
	/* Initialize ClamAV */
	d = avs_clamav_init(
		db_path,
		engine_settings,
		&status
	);
	if(d != NULL)
	{
		mode_t m;
		struct stat stat_buf;
		gint status2;
		gulong directory_level = 0l;
		const gchar *path;
		GList *glist;

		/* Calculate the total number of files to be scanned */
		d->total_files = avs_calculate_total_objects(
			paths_list,
			recursive, executables_only, ignore_links,
			stop_count
		);

		/* Iterate through each object in the paths list */
		for(glist = paths_list; glist != NULL; glist = g_list_next(glist))
		{
			/* Interrupted */
			if(*stop_count > 0)
				break;

			path = (const gchar *)glist->data;
			if(STRISEMPTY(path))
				continue;

			/* Get this object's local stats */
			if(lstat((const char *)path, &stat_buf))
			{
				const gint error_code = (gint)errno;
				avs_print_problem(path, g_strerror(error_code));
				if(status == 0)
					status = -56;
				continue;
			}

			m = stat_buf.st_mode;

			/* Link */
#ifdef S_ISLNK
			if(S_ISLNK(m))
#else
			if(FALSE)
#endif
			{
				if(ignore_links)
				{
					/* Print a warning about this because a link
					 * was explicitly specified and ignore_links
					 * was also specified
					 */
					avs_print_problem(path, "Link ignored");
					continue;
				}

				/* Get this link's destination stats */
				if(stat((const char *)path, &stat_buf))
				{
					const gint error_code = (gint)errno;
					avs_print_problem(path, g_strerror(error_code));
					if(status == 0)
						status = -56;
					continue;
				}

				m = stat_buf.st_mode;

				/* Is this link's destination a directory? */
#ifdef S_ISDIR
				if(S_ISDIR(m))
#else
				if(FALSE)
#endif
				{
					status2 = avs_clamav_scan_directory(
						d,
						path,
						recursive,
						executables_only,
						ignore_links,
						reload,
						stop_count,
						&directory_level,
						engine_settings
					);
				}
				else
				{
					status2 = avs_clamav_scan_file(
						d,
						path,
						stop_count,
						engine_settings
					);
				}
			}
			/* Directory */
#ifdef S_ISDIR
			else if(S_ISDIR(m))
#else
			else if(FALSE)
#endif
				status2 = avs_clamav_scan_directory(
					d,
					path,
					recursive,
					executables_only,
					ignore_links,
					reload,
					stop_count,
					&directory_level,
					engine_settings
				);
			/* Regular File */
#ifdef S_ISREG
			else if(S_ISREG(m))
#else
			else if(TRUE)
#endif
			{
				status2 = avs_clamav_scan_file(
					d,
					path,
					stop_count,
					engine_settings
				);
			}
			else
			{
				/* Unsupported object type or nothing to scan */
				avs_print_problem(path, "Unsupported object type");
				status2 = -52;
			}

			/* If an error occured while scanning this
			 * object and no previous error has occured
			 * then update the status with the new error
			 */
			if((status2 != 0) && (status == 0))
				status = status2;
		}

		avs_clamav_shutdown(d);
	}
	else
	{
		/* Unable to allocate the virus scanner library data */
		if(status == 0)
			status = -1;
	}
#else
	avs_print_error(
		"No AntiVirus support enabled at compile time"
	);
	status = -99;
#endif

	/* User aborted? */
	if(*stop_count > 0)
	{
		if(status == 0)
			status = -4;
	}

	return(status);
}

/*
 *	Front end to execute the AntiVirus Engine to perform a scan.
 *
 *	See avs_avengine_scan() for return values.
 */
gint avs_avengine_scan_main(const gint argc, const gchar **argv)
{
	gboolean	recursive = FALSE,
			executables_only = FALSE,
			ignore_links = FALSE;
	gint		i,
			status;
	gchar *db_path = NULL;
	const gchar	*arg,
			*cwd;
	GList *path_list = NULL;
	AVSAVEngineSettings engine_settings;


#define CLEANUP_RETURN(_v_)	{			\
 /* cl_shutdown(); */					\
							\
 if(path_list != NULL) {				\
  g_list_foreach(path_list, (GFunc)g_free, NULL);	\
  g_list_free(path_list);				\
 }							\
							\
 g_free(db_path);					\
							\
 return(_v_);						\
}

	/* Set the signal callbacks */
	(void)memset(
		&avs_signal_data,
		0x00,
		sizeof(AVSSignalData)
	);
#ifdef SIGINT
	signal(SIGINT, avs_avengine_scan_signal_cb);
#endif
#ifdef SIGTERM
	signal(SIGTERM, avs_avengine_scan_signal_cb);
#endif
#ifdef SIGSEGV
	signal(SIGSEGV, avs_avengine_scan_signal_cb);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE, avs_avengine_scan_signal_cb);
#endif

	/* Get the current working directory */
	cwd = g_get_current_dir();
	if(cwd == NULL)
	   cwd = "/";

	/* Set the default engine settings */
	engine_settings.flags = 0;
	engine_settings.cpu_priority = AVS_AVENGINE_CPU_PRIORITY_HIGH;
	engine_settings.max_scan_size = 0l;
	engine_settings.max_directory_recursions = 0l;
	engine_settings.max_archive_files = 0l;
	engine_settings.max_archive_file_size = 0l;
	engine_settings.max_archive_recursions = 0l;

#if defined(HAVE_CLAMAV)
	/* Initialize ClamAV globals */
	status = (gint)cl_init(CL_INIT_DEFAULT);
	if(status != CL_SUCCESS)
	{
		avs_print_problem("", "Unable to initialize ClamAV");
		CLEANUP_RETURN(-3);
	}
#endif	/* HAVE_CLAMAV */

	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
		arg = (const gchar *)argv[i];
		if(arg == NULL)
			continue;

		/* Scan */
		if(!g_strcasecmp(arg, "--scan"))
		{
			/* Ignore since this was processed in an
			 * earlier encounter which called this function
			 */
		}
		/* Database Path */
		else if(!g_strcasecmp(arg, "--database") ||
			!g_strcasecmp(arg, "-db")
		)
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(arg != NULL)
			{
				db_path = g_strdup(arg);
			}
			else
			{
				g_printerr(
					"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
		}
		/* Recursive */
		else if(!g_strcasecmp(arg, "--recursive") ||
			!g_strcasecmp(arg, "-r")
		)
		{
			recursive = TRUE;
		}
		/* Executables Only */
		else if(!g_strcasecmp(arg, "--executables-only") ||
			!g_strcasecmp(arg, "-x")
		)
		{
			executables_only = TRUE;
		}
		/* Ignore Links */
		else if(!g_strcasecmp(arg, "--ignore-links") ||
			!g_strcasecmp(arg, "-l")
		)
		{
			ignore_links = TRUE;
		}
		/* Report All Objects */
		else if(!g_strcasecmp(arg, "--report-all-objects"))
		{
			engine_settings.flags |= AVS_AVENGINE_REPORT_ALL_OBJECTS;
		}
		/* CPU Priority */
		else if(!g_strcasecmp(arg, "--cpu-priority"))
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(arg != NULL)
			{
				AVSAVEngineCPUPriority p = AVS_AVENGINE_CPU_PRIORITY_LOWEST;
				if(!g_strcasecmp(arg, "highest"))
					p = AVS_AVENGINE_CPU_PRIORITY_HIGHEST;
				else if(!g_strcasecmp(arg, "high"))
					p = AVS_AVENGINE_CPU_PRIORITY_HIGH;
				else if(!g_strcasecmp(arg, "medium"))
					p = AVS_AVENGINE_CPU_PRIORITY_MEDIUM;
				else if(!g_strcasecmp(arg, "low"))
					p = AVS_AVENGINE_CPU_PRIORITY_LOW;
				else if(!g_strcasecmp(arg, "lowest"))
					p = AVS_AVENGINE_CPU_PRIORITY_LOWEST;
				else
				{
					g_printerr(
"%s: Invalid value, valid values are; highest, high, medium, low, lowest.\n",
						argv[i - 1]
					);
					CLEANUP_RETURN(-2);
				}
				engine_settings.cpu_priority = p;
			}
			else
			{
				g_printerr(
					"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
		}
		/* Maximum Scan Size */
		else if(!g_strcasecmp(arg, "--max-scan-size"))
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(arg != NULL)
			{
				engine_settings.max_scan_size = ATOL(arg);
			}
			else
			{
				g_printerr(
					"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
		}
		/* Maximum Directory Recursions */
		else if(!g_strcasecmp(arg, "--max-directory-recursions"))
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(arg != NULL)
			{
				engine_settings.max_directory_recursions = ATOL(arg);
			}
			else
			{
				g_printerr(
					"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
		}
		/* Maximum Archive Files */
		else if(!g_strcasecmp(arg, "--max-archive-files"))
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(arg != NULL)
			{
				engine_settings.max_archive_files = ATOL(arg);
			}
			else
			{
				g_printerr(
					"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
		}
		/* Maximum Archive File Size */
		else if(!g_strcasecmp(arg, "--max-archive-file-size"))
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(arg != NULL)
			{
				engine_settings.max_archive_file_size = ATOL(arg);
			}
			else
			{
				g_printerr(
					"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
		}
		/* Maximum Archive Recursions */
		else if(!g_strcasecmp(arg, "--max-archive-recursions"))
		{
			i++;
			arg = (i < argc) ? argv[i] : NULL;
			if(arg != NULL)
			{
				engine_settings.max_archive_recursions = ATOL(arg);
			}
			else
			{
				g_printerr(
					"%s: Requires argument.\n",
					argv[i - 1]
				);
				CLEANUP_RETURN(-2);
			}
		}
		/* All else assume path */
		else
		{
#ifdef HAVE_LIBENDEAVOUR2
			gchar *path = edv_path_evaluate(
				cwd,
				arg
			);
#else
			gchar *path = avs_join_paths(
				cwd,
				arg
			);
#endif
			if(path != NULL)
				path_list = g_list_append(
					path_list,
					path
				);
		}
	}

	/* Do the scan procedure */
	status = avs_avengine_scan(
		db_path,			/* Database Path */
		path_list,			/* Paths to objects to scan */
		recursive,
		executables_only,
		ignore_links,
		&avs_signal_data.reload,
		&avs_signal_data.stop_count,
		&engine_settings
	);

	CLEANUP_RETURN(status);
#undef CLEANUP_RETURN
}
