#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#if defined(HAVE_CLAMAV)
# include <clamav.h>
#endif
#include <gtk/gtk.h>
#ifdef HAVE_LIBENDEAVOUR2
#include <endeavour2.h>
#endif
#if defined(_WIN32)
# include <windows.h>
#else
# include <unistd.h>
#endif

#include "guiutils.h"
#include "fprompt.h"
#include "cdialog.h"
#include "pdialog.h"
#include "fb.h"

#include "cfg.h"
#include "cfg_fio.h"
#include "avs_avengine_settings.h"
#include "avs_win.h"
#include "avs_win_process.h"
#include "avs_win_list.h"
#include "avs_win_op_cb.h"
#include "avs_core.h"
#include "avs_main.h"

#include "avs_cfg_list.h"
#include "config.h"

#include "images/icon_executable_20x20.xpm"
#include "images/icon_file_infected_20x20.xpm"
#include "images/icon_folder_closed_20x20.xpm"
#include "images/icon_folder_opened_20x20.xpm"
#include "images/icon_virus_pattern_20x20.xpm"


/* Callbacks */
static void avs_main_signal_cb(int s);
static gint avs_main_timeout_cb(gpointer data);

static void avs_main_file_selector_created_cb(
	const gchar *path, gpointer data
);
static void avs_main_file_selector_modified_cb(
	const gchar *old_path,
	const gchar *new_path,
	gpointer data
);
static void avs_main_file_selector_deleted_cb(
	const gchar *path, gpointer data
);

static CfgList *avs_main_new_cfg_list(void);

AVSWin *avs_main_new_avs_win(
	AVSCore *core,
	const gint argc, const gchar **argv
);

gint avs_main(const gint argc, const gchar **argv);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	UNIX signal callback.
 */
static void avs_main_signal_cb(int s)
{
	switch(s)
	{
#ifdef SIGHUP
	    case SIGHUP:
		break;
#endif

#ifdef SIGINT
	    case SIGINT:
		exit(4);		/* User aborted */
		break;
#endif

#ifdef SIGTERM
	    case SIGTERM:
		exit(4);		/* User aborted */
		break;
#endif

#ifdef SIGSEGV
	    case SIGSEGV:
		exit(8);		/* Segmentation fault */
		break;
#endif

#ifdef SIGPIPE
	    case SIGPIPE:
		break;
#endif
	}
}

/*
 *	Main timeout callback.
 */
static gint avs_main_timeout_cb(gpointer data)
{
	gint total_windows = 0;
	GList *glist;
	AVSWin *win;
	AVSCore *core = AVS_CORE(data);
	if(core == NULL)
		return(FALSE);

	/* Manage all windows, count the number of windows that are
	 * mapped and delete any unmapped windows
	 */
	for(glist = core->win;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		win = AVS_WIN(glist->data);
		if(win == NULL)
			continue;

		if(avs_win_is_mapped(win))
		{
			total_windows++;
		}
		else
		{
			avs_win_delete(win);
			glist->data = NULL;
			continue;
		}
	}

	/* If no windows are mapped then break out of the main loop */
	if((total_windows == 0) ||
	   (core->flags & AVS_CORE_CLOSE_ALL_WINDOWS)
	)
	{
		gtk_main_quit();
		core->main_toid = 0;
		return(FALSE);
	}

	return(TRUE);
}


/*
 *	File selector created callback.
 */
static void avs_main_file_selector_created_cb(
	const gchar *path, gpointer data
)
{
#ifdef HAVE_LIBENDEAVOUR2
	AVSCore *core = AVS_CORE(data);
	edv_notify_queue_vfs_object_added(core->edv_ctx, path);
	edv_context_flush(core->edv_ctx);
#endif
}

/*
 *	File selector changed callback.
 */
static void avs_main_file_selector_modified_cb(
	const gchar *old_path,
	const gchar *new_path,
	gpointer data
)
{
#ifdef HAVE_LIBENDEAVOUR2
	AVSCore *core = AVS_CORE(data);
	edv_notify_queue_vfs_object_modified(core->edv_ctx, old_path, new_path);
	edv_context_flush(core->edv_ctx);
#endif
}

/*
 *	File selector deleted callback.
 */
static void avs_main_file_selector_deleted_cb(
	const gchar *path, gpointer data
)
{
#ifdef HAVE_LIBENDEAVOUR2
	AVSCore *core = AVS_CORE(data);
	edv_notify_queue_vfs_object_removed(core->edv_ctx, path);
	edv_context_flush(core->edv_ctx);
#endif
}


/*
 *	Creates a new default Configuration List.
 */
static CfgList *avs_main_new_cfg_list(void)
{
	const CfgList src_cfg_list[] = AVS_CFG_LIST;
	CfgList *cfg_list = CFGItemListCopyList(src_cfg_list);
	const gchar *cwd = g_get_current_dir();
	if(cwd == NULL)
	   cwd = "/";

	CFGItemListSetValueI(
		cfg_list, CFG_PARM_VERSION_MAJOR,
		PROG_VERSION_MAJOR, FALSE
	);
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_VERSION_MINOR,
		PROG_VERSION_MINOR, FALSE
	);
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_VERSION_RELEASE,
		PROG_VERSION_RELEASE, FALSE
	);

	CFGItemListSetValueS(
		cfg_list, CFG_PARM_DB_LOCATION,
		AVS_DEF_DB_LOCATION, FALSE
	);
	CFGItemListSetValueL(
		cfg_list, CFG_PARM_DB_LAST_UPDATE,
		0l, FALSE
	);

	CFGItemListSetValueI(
		cfg_list, CFG_PARM_AVENGINE_SETTING_CPU_PRIORITY,
		AVS_DEF_AVENGINE_SETTING_CPU_PRIORITY, FALSE
	);
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_AVENGINE_SETTING_REPORT_ALL_OBJECTS,
		AVS_DEF_AVENGINE_SETTING_REPORT_ALL_OBJECTS, FALSE
	);
	CFGItemListSetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_SCAN_SIZE,
		AVS_DEF_AVENGINE_SETTING_MAX_SCAN_SIZE, FALSE
	);
	CFGItemListSetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_DIRECTORY_RECURSIONS,
		AVS_DEF_AVENGINE_SETTING_MAX_DIRECTORY_RECURSIONS, FALSE
	);
	CFGItemListSetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_FILES,
		AVS_DEF_AVENGINE_SETTING_MAX_ARCHIVE_FILES, FALSE
	);
	CFGItemListSetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_FILE_SIZE,
		AVS_DEF_AVENGINE_SETTING_MAX_ARCHIVE_FILE_SIZE, FALSE
	);
	CFGItemListSetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_RECURSIONS,
		AVS_DEF_AVENGINE_SETTING_MAX_ARCHIVE_RECURSIONS, FALSE
	);
 
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_WIN_X,
		AVS_DEF_WIN_X, FALSE
	);
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_WIN_Y,
		AVS_DEF_WIN_Y, FALSE
	);
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_WIN_SHOW_TOOL_BAR,
		AVS_DEF_WIN_SHOW_TOOL_BAR, FALSE
	);
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_WIN_SHOW_STATUS_BAR,
		AVS_DEF_WIN_SHOW_STATUS_BAR, FALSE
	);


	return(cfg_list);
}


/*
 *	Creates a new AVSWin.
 */
AVSWin *avs_main_new_avs_win(
	AVSCore *core,
	const gint argc, const gchar **argv
)
{
	gchar *path;
	GList *glist;
	AVSWin *win = avs_win_new(core);
	if(win == NULL)
		return(NULL);

	GUIWindowApplyArgs(
		GTK_WINDOW(win->toplevel),
		argc, (gchar **)argv
	);

	/* Clear the scan settings */
	avs_win_scan_settings_update(win, NULL);

	/* Reset status progress, message, and time */
	avs_win_set_status_progress(win, 0.0f, FALSE);
	avs_win_set_status_message(win, "Ready", FALSE);
	avs_win_set_status_time(win, 0l, FALSE);

	/* Map the window */
	avs_win_map(win);

	avs_win_set_busy(win, TRUE);

	/* Open the scan list from file */
	path = g_strconcat(
		core->prog_data_dir,
		G_DIR_SEPARATOR_S,
		AVS_NAME_SCAN_LIST,
		NULL
	);
	avs_win_scan_list_open_file(
		win,
		path
	);
	g_free(path);

	avs_win_set_busy(win, FALSE);

	/* Append the AVSWin to the list */
	for(glist = core->win;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		if(glist->data == NULL)
		{
			glist->data = win;
			return(win);
		}
	}
	core->win = g_list_append(
		core->win,
		win
	);

	return(win);
}


/*
 *	Runs AVScan normally.
 */
gint avs_main(const gint argc, const gchar **argv)
{
	gboolean	initialized_gtk = FALSE,
			cfg_path_existed;
	gint		i,
			status;
	const gchar	*s,
			*arg,
			*cwd;
	gchar *path;
	GdkWindow *root;
	CfgList *cfg_list;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx;
#endif
	AVSWin *win;
	AVSCore *core;

	/* Set up the signal callbacks */
#ifdef SIGINT
	signal(SIGINT, avs_main_signal_cb);
#endif
#ifdef SIGTERM
	signal(SIGTERM, avs_main_signal_cb);
#endif
#ifdef SIGSEGV
	signal(SIGSEGV, avs_main_signal_cb);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE, avs_main_signal_cb);
#endif

	/* Get the current working directory */
	cwd = g_get_current_dir();
	if(cwd == NULL)
	   cwd = "/";

#ifdef HAVE_LIBENDEAVOUR2
#define DELETE_CORE_LIBS	{			\
 if(core != NULL) {					\
  /* Endeavour Mark II context */			\
  if(core->edv_ctx != NULL) {				\
   edv_context_delete(core->edv_ctx);			\
   core->edv_ctx = NULL;				\
  }							\
 }							\
}
#else
#define DELETE_CORE_LIBS	{ }
#endif

#define CLEANUP_RETURN(_v_)	{			\
							\
 if(core != NULL) {					\
  /* Windows */						\
  if(core->win != NULL) {				\
   g_list_foreach(core->win, (GFunc)avs_win_delete, NULL);	\
   g_list_free(core->win);				\
   core->win = NULL;					\
  }							\
							\
  /* Cursors */						\
  GDK_CURSOR_DESTROY(core->busy_cur);			\
  GDK_CURSOR_DESTROY(core->passive_busy_cur);		\
							\
  /* Pixmaps */						\
  GDK_PIXMAP_UNREF(core->scan_item_pixmap);		\
  GDK_BITMAP_UNREF(core->scan_item_mask);		\
  GDK_PIXMAP_UNREF(core->file_infected_pixmap);		\
  GDK_BITMAP_UNREF(core->file_infected_mask);		\
  GDK_PIXMAP_UNREF(core->folder_closed_pixmap);		\
  GDK_BITMAP_UNREF(core->folder_closed_mask);		\
  GDK_PIXMAP_UNREF(core->folder_opened_pixmap);		\
  GDK_BITMAP_UNREF(core->folder_opened_mask);		\
  GDK_PIXMAP_UNREF(core->db_item_pixmap);		\
  GDK_BITMAP_UNREF(core->db_item_mask);			\
							\
  /* Configuration List */				\
  CFGItemListDeleteList(core->cfg_list);		\
							\
  /* Delete the library handles and resources */	\
  DELETE_CORE_LIBS					\
							\
  g_free(core->prog_name);				\
  g_free(core->prog_path);				\
  g_free(core->prog_data_dir);				\
							\
  g_free(core->title);					\
  g_free(core->geometry);				\
							\
  g_free(core);						\
  core = NULL;						\
 }							\
							\
 if(initialized_gtk) {					\
  /* Shutdown dialogs */				\
  PDialogShutdown();					\
  FileBrowserShutdown();				\
  FPromptShutdown();					\
  CDialogShutdown();					\
 }							\
							\
 return(_v_);						\
}

#ifdef HAVE_CLAMAV
	/* Initialize ClamAV globals */
	status = (gint)cl_init(CL_INIT_DEFAULT);
	if(status != CL_SUCCESS)
	{
		g_printerr(
"Unable to intiialize ClamAV: %s\n",
			cl_strerror((int)status)
		);
		return(1);
	}
#endif

	/* Initialize GTK */
	if(!initialized_gtk)
	{
		if(!gtk_init_check((gint *)&argc, (gchar ***)&argv))
		{
			g_printerr(
"Unable to initialize GTK.\n"
			);
			return(1);
		}

		/* Initialize the GDK RGB buffers */
		gdk_rgb_init();

		/* Get the root GdkWindow */
		root = gdk_window_get_root();

		/* Initialize the dialogs */
		CDialogInit();
		FPromptInit();
		FileBrowserInit();
		PDialogInit();

		initialized_gtk = TRUE;
	}


	/* Create the AVSCore */
	core = AVS_CORE(g_malloc0(sizeof(AVSCore)));
	if(core == NULL)
	{
		CLEANUP_RETURN(-3);
	}

	/* Create a new Configuration List with default values */
	core->cfg_list = cfg_list = avs_main_new_cfg_list();


#ifdef HAVE_LIBENDEAVOUR2
	/* Create a new Endeavour Mark II context */
	core->edv_ctx = edv_ctx = edv_context_new();
#endif

	core->target_scan_item_atom = gdk_atom_intern(
		AVS_DND_TARGET_SCAN_ITEM,
		FALSE
	);

	/* Pixmaps */
	core->scan_item_pixmap = GDK_PIXMAP_NEW_FROM_XPM_DATA(
		&core->scan_item_mask,
		(guint8 **)icon_executable_20x20_xpm
	);
	core->file_infected_pixmap = GDK_PIXMAP_NEW_FROM_XPM_DATA(
		&core->file_infected_mask,
		(guint8 **)icon_file_infected_20x20_xpm
	);
	core->folder_closed_pixmap = GDK_PIXMAP_NEW_FROM_XPM_DATA(
		&core->folder_closed_mask,
		(guint8 **)icon_folder_closed_20x20_xpm
	);
	core->folder_opened_pixmap = GDK_PIXMAP_NEW_FROM_XPM_DATA(
		&core->folder_opened_mask,
		(guint8 **)icon_folder_opened_20x20_xpm
	);
	core->db_item_pixmap = GDK_PIXMAP_NEW_FROM_XPM_DATA(
		&core->db_item_mask,
		(guint8 **)icon_virus_pattern_20x20_xpm
	);

	/* Cursors */
	core->busy_cur = gdk_cursor_new(GDK_WATCH);
	if(TRUE) 
	{
		GdkColor	c_fg = { 0l, 0xffff, 0xffff, 0xffff },
				c_bg = { 0l, 0x0000, 0x0000, 0x0000 };
		const gint width = 24, height = 24;
		const gchar color_mask_data[] = {
0x03, 0x00, 0x00,
0x05, 0x00, 0x00,
0x09, 0x00, 0x00,
0x11, 0x00, 0x00,
0x21, 0x00, 0x00,
0x41, 0x00, 0x00,
0x81, 0x00, 0x00,
0x01, 0x01, 0x00,
0x01, 0xc2, 0x7f,
0xc1, 0x03, 0x00,
0x49, 0x80, 0x3f,
0x95, 0x80, 0x3f,
0x93, 0x80, 0x35,
0x20, 0x01, 0x1b,
0x20, 0x01, 0x0e,
0xc0, 0x00, 0x04,
0x00, 0x00, 0x0a,
0x00, 0x00, 0x1f,
0x00, 0x80, 0x3b,
0x00, 0x80, 0x35,
0x00, 0x80, 0x2a,
0x00, 0x00, 0x00,
0x00, 0xc0, 0x7f,
0x00, 0x00, 0x00
		};
		const gchar trans_mask_data[] = {
0x03, 0x00, 0x00,
0x07, 0x00, 0x00,
0x0f, 0x00, 0x00,
0x1f, 0x00, 0x00,
0x3f, 0x00, 0x00,
0x7f, 0x00, 0x00,
0xff, 0x00, 0x00,
0xff, 0xe1, 0xff,
0xff, 0xe3, 0xff,
0xff, 0xe3, 0xff,
0x7f, 0xc0, 0x7f,
0xf7, 0xc0, 0x7f,
0xf3, 0xc0, 0x7f,
0xe0, 0x81, 0x3f,
0xe0, 0x01, 0x1f,
0xc0, 0x00, 0x0e,
0x00, 0x00, 0x1f,
0x00, 0x80, 0x3f,
0x00, 0xc0, 0x7f,
0x00, 0xc0, 0x7f,
0x00, 0xc0, 0x7f,
0x00, 0xe0, 0xff,
0x00, 0xe0, 0xff,
0x00, 0xe0, 0xff
		};
		GdkBitmap *color_mask = gdk_bitmap_create_from_data(
			root, color_mask_data, width, height
		);
		GdkBitmap *trans_mask = gdk_bitmap_create_from_data(
			root, trans_mask_data, width, height
		);
		core->passive_busy_cur = gdk_cursor_new_from_pixmap(
			color_mask, trans_mask,
			&c_fg, &c_bg,
			0, 0
		);
		GDK_BITMAP_UNREF(color_mask);
		GDK_BITMAP_UNREF(trans_mask);
	}
	else
	{
		core->passive_busy_cur = gdk_cursor_new(GDK_WATCH);
	}

#ifdef HAVE_LIBENDEAVOUR2
	/* Load the Endeavour configuration from file */
	edv_context_init(
		edv_ctx,
		NULL
	);
#endif

	/* Get the full path to this program */
	if(argc > 0)
	{
		arg = argv[0];
		core->prog_name = STRDUP(g_basename(arg));
		core->prog_path = STRDUP(arg);
	}

	/* Get the data directory */
#ifdef HAVE_LIBENDEAVOUR2
	s = edv_get_s(edv_ctx, EDV_CFG_PARM_DIR_LOCAL);
	if(s == NULL)
		s = g_getenv("HOME");
	if(s == NULL)
		s = cwd;
	core->prog_data_dir = STRDUP(s);
#else
	s = g_getenv("HOME");
	if(s == NULL)
		s = cwd;
	core->prog_data_dir = STRDUP(s);
#endif

	/* Open the configuration from file */
	path = g_strconcat(
		core->prog_data_dir,
		G_DIR_SEPARATOR_S,
		AVS_NAME_CONFIGURATION,
		NULL
	);
#ifdef HAVE_LIBENDEAVOUR2
	cfg_path_existed = edv_path_exists(path);
#else
	cfg_path_existed = TRUE;
#endif
	status = CFGFileOpen(
		path,
		cfg_list
	);
	g_free(path);

	/* Set the file selector callbacks */
	FileBrowserSetObjectCreatedCB(
		avs_main_file_selector_created_cb, core
	);
	FileBrowserSetObjectModifiedCB(
		avs_main_file_selector_modified_cb, core
	);
	FileBrowserSetObjectDeletedCB(
		avs_main_file_selector_deleted_cb, core
	);

	/* Create the AVScan window */
	win = avs_main_new_avs_win(
		core,
		argc, argv
	);
	if(win != NULL)
	{
		GList *paths_list = NULL;

#define DELETE_LOCALS	{				\
 if(paths_list != NULL) {				\
  g_list_foreach(paths_list, (GFunc)g_free, NULL);	\
  g_list_free(paths_list);				\
 }							\
}

		/* Reset the current location and scan options */
		avs_win_set_scan_location(win, "");
		avs_win_set_scan_settings(win, 0);

		/* Parse the arguments */
		for(i = 1; i < argc; i++)
		{
			arg = argv[i];
			if(STRISEMPTY(arg))
				continue;

			/* Database Path */
			if(!g_strcasecmp(arg, "--database") ||
			   !g_strcasecmp(arg, "-db")
			)
			{
				i++;
				arg = (i < argc) ? argv[i] : NULL;
				if(arg != NULL)
				{
					CFGItemListSetValueS(
						cfg_list, CFG_PARM_DB_LOCATION,
						arg, FALSE
					);
				}
				else
				{
					g_printerr(
						"%s: Requires argument.\n",
						argv[i - 1]
					);
				}
			}
			/* Recursive */
			else if(!g_strcasecmp(arg, "--recursive") ||
				!g_strcasecmp(arg, "-r")
			)
			{
				gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(win->scan_recursive_check),
					TRUE 
				);
			}
			/* Executables Only */
			else if(!g_strcasecmp(arg, "--executables-only") ||
				!g_strcasecmp(arg, "--executables_only") ||
				!g_strcasecmp(arg, "-x")
			)
			{
				gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(win->scan_executables_only_check),
					TRUE
				);
			}
			/* Ignore Links */
			else if(!g_strcasecmp(arg, "--ignore-links") ||
				!g_strcasecmp(arg, "--ignore_links") ||
				!g_strcasecmp(arg, "-l")
			)
			{
				gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(win->scan_ignore_links_check),
					TRUE
				);
			}
			/* Report All Objects */
			else if(!g_strcasecmp(arg, "--report-all-objects") ||
				!g_strcasecmp(arg, "--report_all_objects")
			)
			{
				CFGItemListSetValueL(
					cfg_list, CFG_PARM_AVENGINE_SETTING_REPORT_ALL_OBJECTS,
					TRUE, FALSE
				);
			}
			/* Do Not Report All Objects */
			else if(!g_strcasecmp(arg, "--no-report-all-objects") ||
				!g_strcasecmp(arg, "--no-report_all_objects")
			)
			{
				CFGItemListSetValueL(
					cfg_list, CFG_PARM_AVENGINE_SETTING_REPORT_ALL_OBJECTS,
					FALSE, FALSE
				);
			}
			/* CPU Priority */
			else if(!g_strcasecmp(arg, "--cpu-priority") ||
				!g_strcasecmp(arg, "--cpu_priority")
			)
			{
				i++;
				arg = (i < argc) ? argv[i] : NULL;
				if(arg != NULL)
				{
					AVSAVEngineCPUPriority p = AVS_AVENGINE_CPU_PRIORITY_LOWEST;
					if(!g_strcasecmp(arg, "highest"))
						p = AVS_AVENGINE_CPU_PRIORITY_HIGHEST;
					else if(!g_strcasecmp(arg, "high"))
						p = AVS_AVENGINE_CPU_PRIORITY_HIGH;
					else if(!g_strcasecmp(arg, "medium"))
						p = AVS_AVENGINE_CPU_PRIORITY_MEDIUM;
					else if(!g_strcasecmp(arg, "low"))
						p = AVS_AVENGINE_CPU_PRIORITY_LOW;
					else if(!g_strcasecmp(arg, "lowest"))
						p = AVS_AVENGINE_CPU_PRIORITY_LOWEST;
					else
					{
						g_printerr(
"%s: Invalid value, valid values are; highest, high, medium, low, lowest.\n",
							argv[i - 1]
						);
						DELETE_LOCALS
						CLEANUP_RETURN(-2);
					}
					CFGItemListSetValueI(
						cfg_list, CFG_PARM_AVENGINE_SETTING_CPU_PRIORITY,
						(gint)p, FALSE
					);
				}
				else
				{
					g_printerr(
						"%s: Requires argument.\n",
						argv[i - 1]
					);
				}
			}
			/* Maximum Scan Size */
			else if(!g_strcasecmp(arg, "--max-scan-size") ||
				!g_strcasecmp(arg, "--max_scan_size")
			)
			{
				i++;
				arg = (i < argc) ? argv[i] : NULL;
				if(arg != NULL)
				{
					CFGItemListSetValueL(
						cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_SCAN_SIZE,
						ATOL(arg), FALSE
					);
				}
				else
				{
					g_printerr(
						"%s: Requires argument.\n",
						argv[i - 1]
					);
				}
			}
			/* Maximum Directory Recursions */
			else if(!g_strcasecmp(arg, "--max-directory-recursions") ||
				!g_strcasecmp(arg, "--max_directory_recursions")
			)
			{
				i++;
				arg = (i < argc) ? argv[i] : NULL;
				if(arg != NULL)
				{
					CFGItemListSetValueL(
						cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_DIRECTORY_RECURSIONS,
						ATOL(arg), FALSE
					);
				}
				else
				{
					g_printerr(
						"%s: Requires argument.\n",
						argv[i - 1]
					);
				}
			}
			/* Maximum Archive Files */
			else if(!g_strcasecmp(arg, "--max-archive-files") ||
				!g_strcasecmp(arg, "--max_archive_files")
			)
			{
				i++;
				arg = (i < argc) ? argv[i] : NULL;
				if(arg != NULL)
				{
					CFGItemListSetValueL(
						cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_FILES,
						ATOL(arg), FALSE
					);
				}
				else
				{
					g_printerr(
						"%s: Requires argument.\n",
						argv[i - 1]
					);
				}
			}
			/* Maximum Archive File Size */
			else if(!g_strcasecmp(arg, "--max-archive-file-size") ||
				!g_strcasecmp(arg, "--max_archive_file_size")
			)
			{
				i++;
				arg = (i < argc) ? argv[i] : NULL;
				if(arg != NULL)
				{
					CFGItemListSetValueL(
						cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_FILE_SIZE,
						ATOL(arg), FALSE
					);
				}
				else
				{
					g_printerr(
						"%s: Requires argument.\n",
						argv[i - 1]
					);
				}
			}
			/* Maximum Archive Recursions */
			else if(!g_strcasecmp(arg, "--max-archive-recursions") ||
				!g_strcasecmp(arg, "--max_archive_recursions")
			)
			{
				i++;
				arg = (i < argc) ? argv[i] : NULL;
				if(arg != NULL)
				{
					CFGItemListSetValueL(
						cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_RECURSIONS,
						ATOL(arg), FALSE
					);
				}
				else
				{
					g_printerr(
						"%s: Requires argument.\n",
						argv[i - 1]
					);
				}
			}
			/* Skip GUI arguments that have an argument */
			else if(!g_strcasecmp(arg, "--geometry") ||
				!g_strcasecmp(arg, "-geometry") ||
				!g_strcasecmp(arg, "--fn") ||
				!g_strcasecmp(arg, "-fn") ||
				!g_strcasecmp(arg, "--fg") ||
				!g_strcasecmp(arg, "-fg") ||
				!g_strcasecmp(arg, "--bg") ||
				!g_strcasecmp(arg, "-bg") ||
				!g_strcasecmp(arg, "--sfg") ||
				!g_strcasecmp(arg, "-sfg") ||
				!g_strcasecmp(arg, "--sbg") ||
				!g_strcasecmp(arg, "-sbg") ||
				!g_strcasecmp(arg, "--cfg") ||
				!g_strcasecmp(arg, "-cfg") ||
				!g_strcasecmp(arg, "--cbg") ||
				!g_strcasecmp(arg, "-cbg") ||
				!g_strcasecmp(arg, "--bg-pixmap") ||
				!g_strcasecmp(arg, "-bg-pixmap") ||
				!g_strcasecmp(arg, "--sbg-pixmap") ||
				!g_strcasecmp(arg, "-sbg-pixmap") ||
				!g_strcasecmp(arg, "--name") ||
				!g_strcasecmp(arg, "-name") ||
				!g_strcasecmp(arg, "--class") ||
				!g_strcasecmp(arg, "-class") ||
				!g_strcasecmp(arg, "--title") ||
				!g_strcasecmp(arg, "-title")
			)
			{
				i++;
			}
			/* All else assume path */
			else if((*arg != '-') && (*arg != '+'))
			{
#ifdef HAVE_LIBENDEAVOUR2
				gchar *path = edv_path_evaluate(
					cwd,
					arg
				);
#else
				gchar *path = g_path_is_absolute(arg) ?
					STRDUP(arg) :
					g_strconcat(cwd, G_DIR_SEPARATOR_S, arg, NULL);
#endif
				if(path != NULL)
					paths_list = g_list_append(
						paths_list,
						path
					);
			}
		}

		/* Need to start scan on startup? */
		if(paths_list != NULL)
		{
			avs_win_start_scan_process(
				win,
				NULL,		/* Use the virus database path
						 * specified in the configuration */
				paths_list,
				avs_win_current_scan_settings(win)
			);

			/* Switch to the results page */
			avs_win_switch_page(win, AVS_WIN_PAGE_RESULTS);
		}
		/* If the configuration file did not exist before then
		 * assume this is the first time running this program
		 */
		else if(!cfg_path_existed)
		{

		}

		DELETE_LOCALS
#undef DELETE_LOCALS
	}

	/* Set the main timeout callback */
	core->main_toid = gtk_timeout_add(
		1000l,				/* Delay in milliseconds */
		avs_main_timeout_cb, core
	);

	/* Enter the main GTK loop */
	gtk_main();


	/* Save the Configuration List to file */
	path = g_strconcat(
		core->prog_data_dir,
		G_DIR_SEPARATOR_S,
		AVS_NAME_CONFIGURATION,
		NULL
	);
	status = CFGFileSave(
		path,
		cfg_list
	);
	g_free(path);

	CLEANUP_RETURN(0);
#undef CLEANUP_RETURN
#undef DELETE_CORE_LIBS
}
