#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>
#if defined(HAVE_CLAMAV)
# include <clamav.h>
#endif
#include <gtk/gtk.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif

#include "cfg.h"
#include "avs_win.h"
#include "avs_win_db.h"
#include "avs_core.h"
#include "avs_cfg_list.h"
#include "config.h"


void avs_win_db_update(
	AVSWin *win,
	const gchar *db_location,
	const gboolean verbose
);
void avs_win_db_queue_update(AVSWin *win);
void avs_win_db_clear(AVSWin *win);

/* Callbacks */
static gint avs_win_db_start_load_idle_cb(gpointer data);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Updates the displayed values on the Virus Database page by
 *	(re)getting the values from the Virus Database.
 */
void avs_win_db_update(
	AVSWin *win,
	const gchar *db_location,
	const gboolean verbose
)
{
#ifdef HAVE_CLAMAV
	struct stat stat_buf;
#endif
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx;
#endif
	GtkWidget *toplevel;
	CfgList *cfg_list;
	AVSCore *core;

	if(win == NULL)
		return;

	toplevel = win->toplevel;
	core = AVS_CORE(win->core);
	cfg_list = core->cfg_list;
#ifdef HAVE_LIBENDEAVOUR2
	edv_ctx = core->edv_ctx;
#endif

	win->stop_count = 0;			/* Actually this function is
						 * not really interruptable */
	avs_win_set_busy(win, TRUE);

	/* Delete the existing virus database and clear the displayed
	 * values
	 */
	avs_win_db_clear(win);

	/* If the virus database location was not specified then use
	 * the one specified in the configuration
	 */
	if(db_location == NULL)
		db_location = CFGItemListGetValueS(
			cfg_list, CFG_PARM_DB_LOCATION
		);
#if defined(HAVE_CLAMAV)
	if(db_location == NULL)
		db_location = cl_retdbdir();
#endif
	if(db_location == NULL)
	{
		if(verbose)
		{
			avs_win_set_status_message(
				win,
"No virus database location specified",
				FALSE
			);
			avs_win_set_status_progress(win, 0.0f, FALSE);
		}
		avs_win_set_busy(win, FALSE);
		return;
	}

	/* Update the displayed Virus Database name and location value */
	gtk_label_set_text(
		GTK_LABEL(win->db_name_label),
		g_basename(db_location)
	);
	gtk_entry_set_text(
		GTK_ENTRY(win->db_location_entry),
		db_location
	);

	/* Set the displayed last updated value */
	if(win->stop_count == 0)
	{
#ifdef HAVE_LIBENDEAVOUR2
		const gulong	cur_t = edv_time(),
				last_updated_t = CFGItemListGetValueL(
					cfg_list, CFG_PARM_DB_LAST_UPDATE
				);
		gchar *s;
		if(last_updated_t > 0l)
		{
			if(cur_t >= last_updated_t)
				s = g_strdup_printf(
					"%s ago",
					edv_date_format_duration(
						edv_ctx,
						cur_t - last_updated_t
					)
				);
			else
				s = g_strdup("(Value is in the future)");
		}
		else
			s = g_strdup("(Unknown)");
		gtk_label_set_text(
			GTK_LABEL(win->db_updated_label),
			s
		);
		g_free(s);
#endif	/* HAVE_LIBENDEAVOUR2 */
	}

#if defined(HAVE_CLAMAV)
	/* Check if the virus database location exists */
	if(win->stop_count == 0)
	{
		if(stat((const char *)db_location, &stat_buf))
		{
			const gint error_code = (gint)errno;
			if(verbose)
			{
				gchar *msg = g_strdup_printf(
"%s: %s.",
					db_location,
					g_strerror(error_code)
				);
				avs_win_set_status_message(win, msg, FALSE);
				g_free(msg);
				avs_win_set_status_progress(win, 0.0f, FALSE);
			}
			avs_win_set_busy(win, FALSE);
			return;
		}

		/* Set the display Virus Database size value (only if
		 * the location was specified as a single file
		 */
		if(S_ISREG(stat_buf.st_mode))
		{
#ifdef HAVE_LIBENDEAVOUR2
			gchar *s = g_strdup_printf(
				"%s bytes",
				edv_str_size_delim((gulong)stat_buf.st_size)
			);
#else
			gchar *s = g_strdup_printf(
				"%ld bytes",
				(gulong)stat_buf.st_size
			);
#endif
			gtk_label_set_text(
				GTK_LABEL(win->db_size_label),
				s
			);
			g_free(s);
		}
		else
		{
			gtk_label_set_text(
				GTK_LABEL(win->db_size_label),
				"-"
			);
		}
	}

	/* CVD */
	if(win->stop_count == 0)
	{
		if(S_ISREG(stat_buf.st_mode))
		{
			struct cl_cvd *cvd = cl_cvdhead(db_location);
			if(cvd != NULL)
			{
				/* Version */
				gchar *s = g_strdup_printf(
					"%ld",
					(gulong)cvd->version
				);
				gtk_label_set_text(
					GTK_LABEL(win->db_version_label),
					s
				);
				g_free(s);

				/* Build Time */
#ifdef HAVE_LIBENDEAVOUR2
				s = STRDUP(edv_date_format_absolute(
					edv_ctx,
					(gulong)cvd->stime
				));
#else
				s = STRDUP(cvd->time);
#endif
				gtk_label_set_text(
					GTK_LABEL(win->db_build_time_label),
					(s != NULL) ? s : ""
				);
				g_free(s);

				/* Patterns */
#ifdef HAVE_LIBENDEAVOUR2
				s = STRDUP(edv_str_size_delim((gulong)cvd->sigs));
#else
				s = g_strdup_printf(
					"%ld",
					(gulong)cvd->sigs
				);
#endif
				gtk_label_set_text(
					GTK_LABEL(win->db_patterns_label),
					s
				);
				g_free(s);

				/* Signature */
				if(cvd->dsig != NULL)
					gtk_entry_set_text(
						GTK_ENTRY(win->db_signature_entry),
						cvd->dsig
					);

				/* MD5 */
				if(cvd->md5 != NULL)
					gtk_entry_set_text(
						GTK_ENTRY(win->db_md5_entry),
						cvd->md5
					);

				/* Builder */
				if(cvd->builder != NULL)
					gtk_entry_set_text(
						GTK_ENTRY(win->db_builder_entry),
						cvd->builder
					);

				/* Functionality Level */
				s = g_strdup_printf(
					"%ld",
					(gulong)cvd->fl
				);
				gtk_label_set_text(
					GTK_LABEL(win->db_functionality_level_label),
					(s != NULL) ? s : ""
				);
				g_free(s);

				cl_cvdfree(cvd);
			}
		}
		else
		{
			gtk_label_set_text(
				GTK_LABEL(win->db_version_label),
				"-"
			);
			gtk_label_set_text(
				GTK_LABEL(win->db_build_time_label),
				"-"
			);
			gtk_label_set_text(
				GTK_LABEL(win->db_patterns_label),
				"-"
			);
			gtk_label_set_text(
				GTK_LABEL(win->db_functionality_level_label),
				"-"
			);
		}
	}
#endif	/* HAVE_CLAMAV */

	/* Resize the GtkLabels due to their values being changed */
	gtk_widget_queue_resize(win->toplevel);

	avs_win_set_busy(win, FALSE);
}


/*
 *	Queues a call to avs_win_db_update().
 */
void avs_win_db_queue_update(AVSWin *win)
{
	if(win == NULL)
		return;

	if(win->db_update_display_idleid != 0)
		return;

	win->db_update_display_idleid = gtk_idle_add_priority(
		G_PRIORITY_LOW,
		avs_win_db_start_load_idle_cb, win
	);
}

/*
 *	Clears all the displayed values for the Virus Database.
 */
void avs_win_db_clear(AVSWin *win)
{
	AVSCore *core;

	if(win == NULL)
		return;

	core = AVS_CORE(win->core);

	gtk_label_set_text(
		GTK_LABEL(win->db_name_label),
		""
	);
	gtk_entry_set_text(
		GTK_ENTRY(win->db_location_entry),
		""
	);
	gtk_label_set_text(
		GTK_LABEL(win->db_updated_label),
		""
	);
	gtk_label_set_text(
		GTK_LABEL(win->db_version_label),
		""
	);
	gtk_label_set_text(
		GTK_LABEL(win->db_build_time_label),
		""
	);
	gtk_label_set_text(
		GTK_LABEL(win->db_patterns_label),
		""
	);
	gtk_label_set_text(
		GTK_LABEL(win->db_size_label),
		""
	);
	gtk_entry_set_text(
		GTK_ENTRY(win->db_signature_entry),
		""
	);
	gtk_entry_set_text(
		GTK_ENTRY(win->db_md5_entry),
		""
	);
	gtk_entry_set_text(
		GTK_ENTRY(win->db_builder_entry),
		""
	);
	gtk_label_set_text(
		GTK_LABEL(win->db_functionality_level_label),
		""
	);

	/* Resize the GtkLabels due to their values being changed */
	gtk_widget_queue_resize(win->toplevel);
}


/*
 *	Virus Database start load idle callback.
 */
static gint avs_win_db_start_load_idle_cb(gpointer data)
{
        AVSWin *win = AVS_WIN(data);
        if(win == NULL)
                return(FALSE);

        win->db_update_display_idleid = 0;

        avs_win_db_update(
                win,
                NULL,				/* Use db location in CfgList */
                TRUE				/* Verbose */
        );
        avs_win_update_display(win);

        return(FALSE);
}
