#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif

#include "../include/string.h"

#include "guiutils.h"

#include "obj.h"
#include "avs_win.h"
#include "avs_win_cb.h"
#include "avs_core.h"
#include "config.h"


gint avs_win_scan_list_insert(
	AVSWin *win,
	const gint row,
	obj_struct *obj				/* Transfered */
);
gint avs_win_scan_list_append(
	AVSWin *win,
	obj_struct *obj				/* Transfered */
);
void avs_win_scan_list_remove(
	AVSWin *win,
	const gint row
);
void avs_win_scan_list_clear(AVSWin *win);
void avs_win_scan_list_update_row(
	AVSWin *win,
	const gint row,
	const obj_struct *obj
);
void avs_win_scan_list_touch_by_location(
	AVSWin *win,
	const gchar *location,
	const gulong t
);

gint avs_win_results_list_append(
	AVSWin *win,
	const gchar *path,
	const gchar *virus_name,
	const gboolean infected
);
void avs_win_results_list_clear(AVSWin *win);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Inserts the Object to the Win's List at the specified row.
 *
 *	The Object will be transfered and should not be referenced
 *	after this call.
 */
gint avs_win_scan_list_insert(
	AVSWin *win,
	const gint row,
	obj_struct *obj                 /* Transfered */
)
{
	gint		i,
			columns,
			new_row;
	gchar **strv;
	GtkCList *clist;

	if(win == NULL)
	{
		ObjDelete(obj);
		return(-1);
	}

	clist = GTK_CLIST(win->scan_clist);
	columns = clist->columns;
	if(columns <= 0)
	{
		ObjDelete(obj);
		return(-1);    
	}

	/* Allocate the row cell values */
	strv = (gchar **)g_malloc(columns * sizeof(gchar *));
	for(i = 0; i < columns; i++)
		strv[i] = "";

	gtk_clist_freeze(clist);

	/* Insert or append a new row */
	if((row >= 0) && (row < clist->rows))
		new_row = gtk_clist_insert(
			clist,
			row,
			strv
		);
	else
		new_row = gtk_clist_append(
			clist,
			strv
		);

	/* Delete the row cell values */
	g_free(strv);

	/* Failed to create new row? */
	if(new_row < 0)
	{
		gtk_clist_thaw(clist);
		ObjDelete(obj);
		return(-1);    
	}

	if(obj != NULL)
	{
		/* Update the row cell values with the new object */
		avs_win_scan_list_update_row(
			win,
			new_row,
			obj
		);

		/* Set the new row's data as the Object */
		gtk_clist_set_row_data_full(
			clist,
			new_row,
			obj, avs_win_scan_list_item_destroy_cb
		);
	}

	gtk_clist_thaw(clist);

	return(new_row);
}

/*
 *	Appends the Object to the Win's List.
 *
 *	The Object will be transfered and should not be referenced
 *	after this call.
 */
gint avs_win_scan_list_append(
	AVSWin *win,
	obj_struct *obj                 /* Transfered */
)
{
	return(avs_win_scan_list_insert(win, -1, obj));
}

/*
 *	Removes the specified row on the Win's List.
 */
void avs_win_scan_list_remove(
	AVSWin *win,
	const gint row
)
{
	GtkCList *clist;

	if(win == NULL)
		return;

	clist = GTK_CLIST(win->scan_clist);
	gtk_clist_freeze(clist);
	gtk_clist_remove(
		clist,
		row
	);
	gtk_clist_thaw(clist);
}

/*
 *	Clears all rows on the Win's List.
 */
void avs_win_scan_list_clear(AVSWin *win)
{
	GtkCList *clist;

	if(win == NULL)
		return;

	clist = GTK_CLIST(win->scan_clist);
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_columns_autosize(clist);
	gtk_clist_thaw(clist);
}

/*
 *	Updates the specified row on the Win's List with the values
 *	specified in the Object.
 */
void avs_win_scan_list_update_row(
	AVSWin *win,
	const gint row,
	const obj_struct *obj
)
{
	gint		npaths,
			column,
			columns;
#ifdef HAVE_LIBENDEAVOUR2
	const gulong cur_time = edv_time();
#endif
	gchar *path;
	GdkPixmap	*location_pixmap = NULL,
			*name_pixmap = NULL;
	GdkBitmap	*location_mask = NULL,
			*name_mask = NULL;
	GtkCList *clist;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *ctx;
#endif
	AVSCore *core;

	if(win == NULL)
		return;

	clist = GTK_CLIST(win->scan_clist);
	columns = clist->columns;
	if((row < 0) || (row >= clist->rows) || (columns <= 0))
		return;

	core = AVS_CORE(win->core);
#ifdef HAVE_LIBENDEAVOUR2
	ctx = core->edv_ctx;
#endif

	/* Get the scan item pixmap */
	name_pixmap = core->scan_item_pixmap;
	name_mask = core->scan_item_mask;
	GDK_PIXMAP_REF(name_pixmap);
	GDK_BITMAP_REF(name_mask);

	/* Get a copy of the first path if specified and the number of
	 * paths specified
	 */
	if(obj != NULL)
	{
		if(obj->path != NULL)
		{
			if(strchr((char *)obj->path, ',') != NULL)
			{
				gint i;
				gchar **strv = g_strsplit(obj->path, ",", -1);

				npaths = 0;
				for(i = 0; strv[i] != NULL; i++)
					npaths++;

				path = (npaths > 0) ? STRDUP(strv[0]) : NULL;

				g_strfreev(strv);
			}
			else
			{
				path = STRDUP(obj->path);
				npaths = 1;
			}
		}
		else
		{
			path = NULL;
			npaths = 0;
		}
	}
	else
	{
		path = NULL;
		npaths = 0;
	}

#ifdef HAVE_LIBENDEAVOUR2
	/* Match the icon for the location */
	if(!STRISEMPTY(path))
	{
		EDVVFSObject *obj = edv_vfs_object_lstat(path);
		if(obj != NULL)
		{
			gchar	*icon_path,
					*icon_opened_path,
					*icon_inaccessable_path,
					*icon_hidden_path;
			(void)edv_match_vfs_object_icon(
				ctx,
				obj,
				EDV_ICON_SIZE_20,
				&icon_path,
				&icon_opened_path,
				&icon_inaccessable_path,
				&icon_hidden_path
			);
			if(edv_path_is_hidden(path))
			{
				if(icon_hidden_path != NULL)
				{
					g_free(icon_path);
					icon_path = g_strdup(icon_hidden_path);
				}
			}
			if(icon_path != NULL)
				location_pixmap = GDK_PIXMAP_NEW_FROM_XPM_FILE(
					&location_mask,
					icon_path
				);
			g_free(icon_path);
			g_free(icon_opened_path);
			g_free(icon_inaccessable_path);
			g_free(icon_hidden_path);
			edv_vfs_object_delete(obj);
		}
	}
#endif	/* HAVE_LIBENDEAVOUR2 */


	/* Begin updating the cells */
	gtk_clist_freeze(clist);

	/* Name */
	column = 0;
	if(column < columns)
	{
		gchar *s = avs_win_strdup_shorten(
			(obj != NULL) ? obj->name : NULL,
			20
		);
		if(name_pixmap != NULL)
			gtk_clist_set_pixtext(
				clist, row, column,
				(s != NULL) ? s : "",
				AVS_WIN_LIST_PIXMAP_TEXT_SPACING,
				name_pixmap, name_mask
			);
		else
			gtk_clist_set_text(
				clist, row, column,
				(s != NULL) ? s : ""
			);
		g_free(s);
	}

	/* Location */
	column = 1;
	if(column < columns)
	{
		gchar *label;

		if(path != NULL)
		{
			if(npaths > 1)
			{
				gchar *s;
				if(npaths > 2)
					s = g_strdup_printf(
						"%s, (+%i...)",
						path,
						npaths - 1
					);
				else
					s = g_strconcat(
						path,
						", (...)",
						NULL
					);
				label = avs_win_strdup_shorten(s, 25);
				g_free(s);
			}
			else
			{
				label = avs_win_strdup_shorten(path, 25);
			}
		}
		else
		{
			label = NULL;
		}

		if(location_pixmap != NULL)
			gtk_clist_set_pixtext(
				clist, row, column,
				(label != NULL) ? label : "",
				AVS_WIN_LIST_PIXMAP_TEXT_SPACING,
				location_pixmap, location_mask
			);
		else  
		    gtk_clist_set_text(
				clist, row, column,
				(label != NULL) ? label : ""
		    );
		g_free(label);
	}

	/* Last Runned */
	column = 2;
	if(column < columns)
	{
#ifdef HAVE_LIBENDEAVOUR2
		const gchar *s = NULL;
		const gulong t = (obj != NULL) ? obj->last_runned : 0l;
		if(t > 0l)
		{
			const gulong dt = (cur_time > t) ? (cur_time - t) : 0l;
			/* Less than 6 months ago? */
			if(dt < (6 * 30 * 24 * 60 * 60))
				s = edv_date_format_current(ctx, t);
			else
				s = edv_date_format_absolute(ctx, t);
		}
		if(s == NULL)
			s = "Unknown";

		gtk_clist_set_text(clist, row, column, s);
#else
		gchar buf[80], *s2;
		const gulong t = (obj != NULL) ? obj->last_runned : 0l;
		time_t tt = (time_t)t;
		strncpy(buf, ctime(&tt), sizeof(buf));
		s2 = (gchar *)strchr((char *)buf, '\n');
		if(s2 != NULL)
			*s2 = '\0';
		gtk_clist_set_text(clist, row, column, buf);
#endif
	}     

	g_free(path);

	gtk_clist_columns_autosize(clist);

	gtk_clist_thaw(clist);

	GDK_PIXMAP_UNREF(name_pixmap);
	GDK_BITMAP_UNREF(name_mask);
	GDK_PIXMAP_UNREF(location_pixmap);
	GDK_BITMAP_UNREF(location_mask);
}

/*
 *	Updates the last runned time of all scan items who's
 *	location matches the specified location.
 */
void avs_win_scan_list_touch_by_location(
	AVSWin *win,
	const gchar *location,
	const gulong t
)
{
	gchar **locations_list;
	GtkCList *clist;
	obj_struct *obj;

	if((win == NULL) || STRISEMPTY(location))
		return;

	clist = GTK_CLIST(win->scan_clist);

	locations_list = g_strsplit(location, ",", -1);
	if(locations_list != NULL)
	{
		const gint nrows = clist->rows;
		gint	i,
					row;

		for(i = 0; locations_list[i] != NULL; i++)
		{
			for(row = 0; row < nrows; row++)
			{
				obj = OBJ(gtk_clist_get_row_data(
					clist,
					row
				));
				if(obj == NULL)
					continue;

				if(obj->path == NULL)
					continue;

				if(strstr(
					(const char *)obj->path,
					(const char *)locations_list[i]
				) != NULL)
				{
					obj->last_runned = t;
					avs_win_scan_list_update_row(
						win,
						row,
						obj
					);
				}
			}
		}

		g_strfreev(locations_list);
	}
}


/*
 *	Appends the row to the Win's Results List.
 */
gint avs_win_results_list_append(
	AVSWin *win,
	const gchar *path, const gchar *virus_name,
	const gboolean infected
)
{
	gint i, column, row, new_row, columns;
	gchar **strv;
	GdkPixmap *pixmap = NULL;
	GdkBitmap *mask = NULL;
	GtkCList *clist;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *ctx;
#endif
	obj_struct *obj;
	AVSCore *core;

	if(win == NULL)
		return(-1);

	clist = GTK_CLIST(win->results_clist);
	columns = clist->columns;
	if(columns <= 0)
		return(-1);

	core = AVS_CORE(win->core);
#ifdef HAVE_LIBENDEAVOUR2
	ctx = core->edv_ctx;
#endif

	/* Allocate the row cell values */
	strv = (gchar **)g_malloc(columns * sizeof(gchar *));
	for(i = 0; i < columns; i++)                         
		strv[i] = "";

	gtk_clist_freeze(clist);

	/* Append a new row */
	new_row = gtk_clist_append(clist, strv);

	/* Delete row cell values */
	g_free(strv);

	/* Failed to create new row? */
	if(new_row < 0)                
	{
		gtk_clist_thaw(clist);
		return(-1);
	}

	/* Create an object to hold the result information and set it
	 * as the new row's data
	 */
	obj = ObjNew();
	if(obj != NULL)
	{
		obj->path = STRDUP(path);
		if(infected)
			obj->options |= AVS_SCAN_INFECTED;

		/* Set the new row's data as the Object */
		gtk_clist_set_row_data_full(
			clist, new_row,
			obj, avs_win_results_list_item_destroy_cb
		);
	}

	/* Infected with virus? */
	if(infected)
	{
		gtk_clist_set_row_style(clist, new_row, win->infected_style);
		pixmap = core->file_infected_pixmap;
		mask = core->file_infected_mask;
		GDK_PIXMAP_REF(pixmap);
		GDK_BITMAP_REF(mask);
	}
	/* Match Device or MIME Type */
	else if(!STRISEMPTY(path))
	{
#ifdef HAVE_LIBENDEAVOUR2
		EDVVFSObject *obj = edv_vfs_object_lstat(path);
		if(obj != NULL)
		{
			gchar	*icon_path,
					*icon_opened_path,
					*icon_inaccessable_path,
					*icon_hidden_path;
			(void)edv_match_vfs_object_icon(
				ctx,
				obj,
				EDV_ICON_SIZE_20,
				&icon_path,
				&icon_opened_path,
				&icon_inaccessable_path,
				&icon_hidden_path
			);
			if(edv_path_is_hidden(path))
			{
				if(icon_hidden_path != NULL)
				{
					g_free(icon_path);
					icon_path = g_strdup(icon_hidden_path);
				}
			}
			if(icon_path != NULL)
				pixmap = GDK_PIXMAP_NEW_FROM_XPM_FILE(
					&mask,
					icon_path
				);
			g_free(icon_path);
			g_free(icon_opened_path);
			g_free(icon_inaccessable_path);
			g_free(icon_hidden_path);
			edv_vfs_object_delete(obj);
		}
#endif	/* HAVE_LIBENDEAVOUR2 */
	}


	/* Begin setting cells */
	row = new_row;

	/* Object */
	column = 0;
	if(column < columns)
	{
/*	    gchar *s = avs_win_strdup_shorten(path, 32); */
		const gchar *s = path;
		if(pixmap != NULL)
			gtk_clist_set_pixtext(
				clist, row, column,
				(s != NULL) ? s : "",
				AVS_WIN_LIST_PIXMAP_TEXT_SPACING,
				pixmap, mask 
			);
		else
			gtk_clist_set_text(  
				clist, row, column,
				(s != NULL) ? s : ""
			);
/*	    g_free(s); */
	}     

	/* Virus/Problem */
	column = 1;
	if(column < columns)
	{
		const gchar *s = virus_name;
		gtk_clist_set_text(
			clist, row, column,
			(s != NULL) ? s : ""   
		);
	}

	gtk_clist_columns_autosize(clist);

	gtk_clist_thaw(clist);

	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);

	return(new_row);
}


/*
 *      Clears all rows on the Win's Results List.
 */
void avs_win_results_list_clear(AVSWin *win)
{
	GtkCList *clist;

	if(win == NULL)
		return;

	clist = GTK_CLIST(win->results_clist);
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_columns_autosize(clist);
	gtk_clist_thaw(clist);
} 
