#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif
#if defined(_WIN32)
# include <windows.h>
#else
# include <unistd.h>
#endif

#ifndef HAVE_LIBENDEAVOUR2
# include "../include/prochandle.h"
#endif

#include "cfg.h"

#include "animicon.h"
#include "cdialog.h"
#include "avs_avengine_settings.h"
#include "avs_win.h"
#include "avs_win_process.h"
#include "avs_win_animicon.h"
#include "avs_win_list.h"
#include "avs_win_db.h"
#include "avs_win_op_cb.h"
#include "avs_core.h"

#include "avs_cfg_list.h"
#include "config.h"


/* Execute */
static gint avs_win_process_execute(
	AVSWin *win,
	const gchar *cmd
);

/* Delete Process Output */
void avs_win_delete_process_output(AVSWin *win);

/* Scan Process */
void avs_win_start_scan_process(
	AVSWin *win,
	const gchar *db_location,
	GList *paths_list,
	const AVSScanSettings options
);
void avs_win_stop_scan_process(AVSWin *win);
gboolean avs_win_scan_process_is_running(AVSWin *win);
static gint avs_win_scan_monitor_timeout_cb(gpointer data);

/* Virus Database Update Process */
void avs_win_start_db_update_net(AVSWin *win);
void avs_win_stop_db_update_net(AVSWin *win);
gboolean avs_win_db_update_net_process_is_running(AVSWin *win);
static gint avs_win_db_update_monitor_timeout_cb(gpointer data);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#define UNLINK(p)	(((p) != NULL) ? (gint)unlink((const char *)p) : -1)
#define FCLOSE(p)	(((p) != NULL) ? (gint)fclose(p) : -1)

#define ISBLANK(c)	(((c) == ' ') || ((c) == '\t'))


/*
 *	Executes the command and opens the output stream to the AVSWin.
 *
 *	Note that avs_win_delete_process_output() should be called
 *	prior to this function so that the streams can be closed and
 *	output files deleted from the AVSWin.
 *
 *	Returns the process ID on success or 0 or negative on error.
 */
static gint avs_win_process_execute(
	AVSWin *win,
	const gchar *cmd
)
{
#ifdef HAVE_LIBENDEAVOUR2
	gchar *shell_prog;
	AVSCore *core = win->core;
	EDVContext *edv_ctx = core->edv_ctx;

	/* Get the configured shell program and arguments */
	const gchar	*shell_cmd = edv_get_s(edv_ctx, EDV_CFG_PARM_PROG_SHELL),
			*shell_args = edv_strarg(
		shell_cmd,
		&shell_prog,
		TRUE,				/* Parse escapes */
		TRUE				/* Parse quotes */
	);

	/* Execute the command */
	const gint pid = edv_system_shell_streams(
		cmd,
		shell_prog,
		shell_args,
		NULL,
		&win->cstdout,
		&win->cstderr
	);

	g_free(shell_prog);

	return(pid);
#else
	gint		fd,
			pid;

	/* Get the path to the temporary files directory */
	const gchar *tmp_dir = g_getenv(ENV_VAR_NAME_TMPDIR);
	if(tmp_dir == NULL)
#if defined(P_tmpdir)
		tmp_dir = P_tmpdir;
#elif defined(_WIN32)
		tmp_dir = "C:\\TEMP";
#else
		tmp_dir = "/tmp";
#endif

	/* Format the standard output file path */
	win->cstdout_path = g_strconcat(
		tmp_dir,
		G_DIR_SEPARATOR_S,
		PROG_NAME,
		"XXXXXX",
		NULL
	);
#ifdef _WIN32
	fd = -1;
	mktemp((char *)win->cstdout_path);
#else
	fd = (gint)mkstemp((char *)win->cstdout_path);
	if(fd > -1)
		(void)close((int)fd);
#endif

	/* Format the standard error file path */
	win->cstderr_path = g_strconcat(
		tmp_dir,
		G_DIR_SEPARATOR_S,
		PROG_NAME,
		"XXXXXX",
		NULL
	);
#ifdef _WIN32
	fd = -1;
	mktemp((char *)win->cstderr_path);
#else
	fd = (gint)mkstemp((char *)win->cstderr_path);
	if(fd > -1)
		(void)close((int)fd);
#endif

	/* Execute the command */
	pid = (gint)ExecOE(
		cmd,
		win->cstdout_path,
		win->cstderr_path
	);
	if(pid > 0)
	{
		win->cstdout = fopen((const char *)win->cstdout_path, "rb");
		win->cstderr = fopen((const char *)win->cstderr_path, "rb");
	}

	return(pid);
#endif	/* HAVE_LIBENDEAVOUR2 */
}


/*
 *	Closes the streams to the child process and removes any output
 *	files.
 */
void avs_win_delete_process_output(AVSWin *win)
{
	if(win == NULL)
		return;

	/* Close and remove the stdout & stderr files */
	(void)FCLOSE(win->cstdout);
	win->cstdout = NULL;

	g_free(win->cstdout_line_buf);
	win->cstdout_line_buf = NULL;

	(void)UNLINK(win->cstdout_path);
	g_free(win->cstdout_path);
	win->cstdout_path = NULL;

	(void)FCLOSE(win->cstderr);
	win->cstderr = NULL;

	g_free(win->cstderr_line_buf);
	win->cstderr_line_buf = NULL;

	(void)UNLINK(win->cstderr_path);
	g_free(win->cstderr_path);
	win->cstderr_path = NULL;
}

/*
 *	Starts the scan process.
 *
 *	Any current scan process will be stopped first.
 *
 *	The db_location specifies the full path to the Virus Scanner
 *	Database. If this value is NULL then the CFG_PARM_DB_LOCATION
 *	value from the configuration will be used instead.
 *
 *	The paths_list specifies the list of full paths to the locations
 *	to scan.
 *
 *	The options specifies the scan options.
 */
void avs_win_start_scan_process(
	AVSWin *win,
	const gchar *db_location,
	GList *paths_list,
	const AVSScanSettings options
)
{
	gboolean report_all_objects;
	gint pid;
	gulong		max_scan_size,
			max_directory_recursions,
			max_archive_files,
			max_archive_file_size,
			max_archive_recursions;
	const gchar	*cpu_priority_s,
			*path;
	gchar		*cmd,
			*cmd_next;
#ifdef HAVE_LIBENDEAVOUR2
	const gulong start_time = edv_time();
	EDVContext *edv_ctx;
#else
	const gulong start_time = (gulong)time(NULL);
#endif
	GList *glist;
	CfgList *cfg_list;
	GtkWidget *toplevel;
	AVSWinScanContext *scan_ctx;
	AVSCore *core;

	if(win == NULL)
		return;

	toplevel = win->toplevel;
	scan_ctx = win->scan_context;
	core = AVS_CORE(win->core);
	cfg_list = core->cfg_list;
#ifdef HAVE_LIBENDEAVOUR2
	edv_ctx = core->edv_ctx;
#endif

	/* Clear the status message */
	avs_win_set_status_message(win, NULL, FALSE);

	/* Stop any existing scan process, close any streams to it,
	 * and remove any output files
	 */
	avs_win_stop_scan_process(win);

	/* Path to this program not known? */
	if(STRISEMPTY(core->prog_path))
	{
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_error(edv_ctx); 
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Scan Failed",
"Internal error:\n\
\n\
Unable to obtain the path to this process.",
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		avs_win_set_results_status(win, "Failed");
		return;
	}

	/* Get the virus database path from the configuration if it
	 * was not specified
	 */
	if(db_location == NULL)
		db_location = CFGItemListGetValueS(
			cfg_list, CFG_PARM_DB_LOCATION
		);
	if(STRISEMPTY(db_location))
	{
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_warning(edv_ctx); 
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Scan Failed",
"The virus database location was not specified.\n\
\n\
To set the virus database location go to\n\
Database->Location...",
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		avs_win_set_results_status(win, "Failed");
		return;
	}

#if defined(HAVE_CLAMAV) && defined(HAVE_LIBENDEAVOUR2)
	/* Invalid virus database location? */
	if(!edv_path_exists(db_location))
	{
		gchar *msg = g_strdup_printf(
"The specified virus database location does not exist:\n\
\n\
    %s\n\
\n\
To set the virus database location go to\n\
Database->Location...",
			db_location
		);
		edv_play_sound_warning(edv_ctx); 
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Scan Failed",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		g_free(msg);
		CDialogSetTransientFor(NULL);
		avs_win_set_results_status(win, "Failed");
		return;
	}
#endif	/* HAVE_CLAMAV && HAVE_LIBENDEAVOUR2 */

	/* Nothing to scan? */
	if(paths_list == NULL)
	{
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_warning(edv_ctx);
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Scan Failed",
"The scan location has not been specified.",
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		avs_win_set_results_status(win, "Failed");
		return;
	}

	/* Get the current AVSAVEngineSettings */
	switch((AVSAVEngineCPUPriority)CFGItemListGetValueI(
		cfg_list, CFG_PARM_AVENGINE_SETTING_CPU_PRIORITY
	))
	{
	    case AVS_AVENGINE_CPU_PRIORITY_HIGHEST:
		cpu_priority_s = "highest";
		break;
	    case AVS_AVENGINE_CPU_PRIORITY_HIGH:
		cpu_priority_s = "high";
		break;
	    case AVS_AVENGINE_CPU_PRIORITY_MEDIUM:
		cpu_priority_s = "medium";
		break;
	    case AVS_AVENGINE_CPU_PRIORITY_LOW:
		cpu_priority_s = "low";
		break;
	    case AVS_AVENGINE_CPU_PRIORITY_LOWEST:
		cpu_priority_s = "lowest";
		break;
	    default:
		cpu_priority_s = "highest";
		break;
	}
	report_all_objects = (gboolean)CFGItemListGetValueI(
		cfg_list, CFG_PARM_AVENGINE_SETTING_REPORT_ALL_OBJECTS
	);
	max_scan_size = CFGItemListGetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_SCAN_SIZE
	);
	max_directory_recursions = CFGItemListGetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_DIRECTORY_RECURSIONS
	);
	max_archive_files = CFGItemListGetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_FILES
	);
	max_archive_file_size = CFGItemListGetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_FILE_SIZE
	);
	max_archive_recursions = CFGItemListGetValueL(
		cfg_list, CFG_PARM_AVENGINE_SETTING_MAX_ARCHIVE_RECURSIONS
	);

	/* Set the status message */
	avs_win_set_status_message(
		win,
		"Scan in progress...",
		FALSE
	);

	/* Set the results location and status message */
	avs_win_set_results_location(
		win,
		paths_list
	);
	avs_win_set_results_status(
		win,
		"Scan in progress..."
	);

	/* Clear the Results List */
	avs_win_results_list_clear(win);

	/* Start playing the animated icon */
	AnimIconPlay(win->anim_icon);

	/* Set the scan monitor timeout callback and increment the
	 * passively busy count based on the setting of the timeout
	 * callback
	 */
	scan_ctx->toid = gtk_timeout_add(
		250l,
		avs_win_scan_monitor_timeout_cb, win
	);
	if(scan_ctx->toid != 0)
		avs_win_set_passive_busy(win, TRUE);

	/* Format the scan command */
	cmd = g_strdup_printf(
"\"%s\"\
 --scan%s%s%s\
 --database \"%s\"\
 --cpu-priority \"%s\"\
%s\
 --max-scan-size %ld\
 --max-directory-recursions %ld\
 --max-archive-files %ld\
 --max-archive-file-size %ld\
 --max-archive-recursions %ld",
		core->prog_path,
		(options & AVS_SCAN_RECURSIVE) ?
			" --recursive" : "",
		(options & AVS_SCAN_EXECUTABLES_ONLY) ?
			" --executables-only" : "",
		(options & AVS_SCAN_IGNORE_LINKS) ?
			" --ignore-links" : "",
		db_location,
		cpu_priority_s,
		report_all_objects ? " --report-all-objects" : "",
		max_scan_size,
		max_directory_recursions,
		max_archive_files,
		max_archive_file_size,
		max_archive_recursions
	);
	for(glist = paths_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		path = (const gchar *)glist->data;
		if(STRISEMPTY(path))
			continue;

		cmd_next = g_strconcat(
			cmd,
			" \"",
			path,
			"\"",
			NULL
		);
		if(cmd_next == NULL)
		{
			g_free(cmd);
			cmd = NULL;
			break;
		}
		g_free(cmd);
		cmd = cmd_next;
	}
	/* Unable to generate the scan command? */
	if(cmd == NULL)
	{
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_error(edv_ctx);
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Scan Failed",
			"Unable to generate the scan command.",
			NULL,
			CDIALOG_ICON_ERROR,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);

		avs_win_stop_scan_process(win);
		avs_win_set_results_status(win, "Failed");
		avs_win_update_display(win);
		return;
	}

	/* Execute the scan command, open the streams to the new
	 * process and store them on the AVSWin
	 */
	scan_ctx->pid = pid = avs_win_process_execute(
		win,
		cmd
	);
	if(pid <= 0)
	{
		const gint error_code = (gint)errno;
		char *msg = g_strdup_printf(
"Unable to execute:\n\
\n\
    %s\n\
\n\
%s.",
			cmd,
			g_strerror(error_code)
		);
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_error(edv_ctx);
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Scan Failed",
			msg,
			NULL,
			CDIALOG_ICON_ERROR,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		free(msg);
		CDialogSetTransientFor(NULL);

		avs_win_stop_scan_process(win);
		avs_win_set_results_status(win, "Failed");
		avs_win_update_display(win);

		g_free(cmd);
		return;
	}

	/* Delete the scan command */
	g_free(cmd);

	/* Record the start time */
	scan_ctx->start_time = start_time;
	scan_ctx->duration = 0l;

	/* Reset the scan stats */
	scan_ctx->nscanned = 0l;
	scan_ctx->ninfected = 0l;
	scan_ctx->nproblems = 0l;
	scan_ctx->total = 0l;
	scan_ctx->blocks_scanned = 0l;

	/* Reset the progress and time */
	avs_win_set_status_progress(win, 0.0f, FALSE);
	avs_win_set_status_time(win, 0l, FALSE);

	avs_win_update_title(win);
	avs_win_update_display(win);
}

/*
 *	Stops the scanning process and cleans up.
 *
 *	Stops the scan icon animation.
 *
 *	If the scan monitoring timeout was set then it will be removed
 *	and decrements a passive busy count.
 *
 *	Interruptes the scan process, closes any streams and deletes
 *	any output files.
 *
 *	Resets the status bar and scan values.
 */
void avs_win_stop_scan_process(AVSWin *win)
{
	gboolean was_monitoring_process;
	AVSWinScanContext *scan_ctx;

	if(win == NULL)
		return;

	scan_ctx = win->scan_context;

	/* Stop animated icon from playing */
	AnimIconPause(win->anim_icon);
	AnimIconSeek(win->anim_icon, 0);

	/* Remove the scan timeout callback */
	if(scan_ctx->toid != 0)
	{
		gtk_timeout_remove(scan_ctx->toid);
		scan_ctx->toid = 0;
		was_monitoring_process = TRUE;
	}
	else
		was_monitoring_process = FALSE;

	/* Interrupt the scan (child) process
	 *
	 * Sending SIGINT to the scan process will increment its stop
	 * count in the signal callback but it will not exit()
	 * immediately, instead it will wait for the current (if any)
	 * ClamAV call to finish and continue to execute without calling
	 * any additional ClamAV calls except AVScanOPClamAVShutdown()
	 * in order to shut down ClamAV nicely (so that there are no
	 * ClamAV tempory files) and finish execution
	 */
	if(scan_ctx->pid > 0)
	{
#ifdef SIGCONT
		if(win->paused)
			(void)kill(
				(int)scan_ctx->pid,
				SIGCONT
			);
#endif
#ifdef SIGINT
		(void)kill(
			(int)scan_ctx->pid,
			SIGINT
		);
#else
# warning "SIGINT not defined, scan process stopping will not function."
#endif
		scan_ctx->pid = 0;
	}

	/* Close the child streams */
	avs_win_delete_process_output(win);

	/* Reset the stop counter and paused state */
	win->stop_count = 0;
	win->paused = FALSE;

	/* Reset the progress and time */
	avs_win_set_status_progress(win, 0.0f, FALSE);
	avs_win_set_status_time(win, 0l, FALSE);

	avs_win_update_title(win);
	avs_win_update_display(win);

	/* Mark the AVSWin as no longer passively busy */
	if(was_monitoring_process)
		avs_win_set_passive_busy(win, FALSE);
}

/*
 *	Checks if the scan process is running.
 */
gboolean avs_win_scan_process_is_running(AVSWin *win)
{
	if(win == NULL)
		return(FALSE);

	if(win->scan_context->pid <= 0)
		return(FALSE);

#ifdef HAVE_LIBENDEAVOUR2
	return(edv_pid_exists(win->scan_context->pid));
#else
	return(ExecProcessExists((pid_t)win->scan_context->pid));
#endif
}

/*
 *	Scan process monitoring timeout callback.
 *
 *	Monitors the scan process and handles/displays any messages from
 *	it.
 */
static gint avs_win_scan_monitor_timeout_cb(gpointer data)
{
	gint pid;
	FILE *fp;
	GtkWidget *toplevel;
	GtkCList *clist;
	AVSCore *core;
	AVSWinScanContext *scan_ctx;
	AVSWin *win = AVS_WIN(data);
	if(win == NULL)
		return(FALSE);

	toplevel = win->toplevel;
	clist = (GtkCList *)win->results_clist;
	core = AVS_CORE(win->core);
	scan_ctx = win->scan_context;
	pid = scan_ctx->pid;

	/* User clicked on the stop button? */
	if(win->stop_count > 0)
	{
		/* Stop scanning
		 *
		 * Mark that this timeout is no longer being called and
		 * decrement the passive busy count
		 */
		scan_ctx->toid = 0;
		avs_win_set_passive_busy(win, FALSE);

		/* Close the streams and delete any output files */
		avs_win_stop_scan_process(win);

		avs_win_set_results_status(
			win,
			"Scan Interrupted"
		);
		avs_win_set_status_message(
			win,
"Scan interrupted",
			FALSE
		);

		/* Do not call this function again */
		return(FALSE);
	}

	/* Check for any handle any messages from stderr */
	fp = win->cstderr;
	if(fp != NULL)
	{
		/* Anything from stderr will be added as a string to
		 * the virus/problem column of the results list
		 */
		gboolean got_line = FALSE;
#ifdef HAVE_LIBENDEAVOUR2
		const gint fd = (gint)fileno(fp);
#endif
		gint new_row;
		gchar	**line_buf = &win->cstderr_line_buf,
			*s,
			buf[32];
#ifdef HAVE_LIBENDEAVOUR2
		while(edv_poll_read(fd))
#else
		while(TRUE)
#endif
		{
			if(fgets((char *)buf, sizeof(buf), fp) == NULL)
				break;

			if(*line_buf != NULL)
				s = g_strconcat(
					*line_buf,
					buf,
					NULL
				);
			else
				s = g_strdup(buf);
			if(s != NULL)
			{
				g_free(*line_buf);
				*line_buf = s;
				s = (gchar *)strpbrk((char *)s, "\n\r");
				if(s != NULL)
				{
					*s = '\0';
					got_line = TRUE;
				}
			}
			if(got_line)
			{
				/* Seek past any spaces */
				const gchar *s = *line_buf;
				while(ISBLANK(*s))
					s++;

				new_row = avs_win_results_list_append(
					win,
					NULL,
					s,
					FALSE
				);
				if(new_row > -1)
				{
					gtk_clist_moveto(
						clist,
						new_row, 0,
						1.0f, 0.0f
					);
				}
	
				g_free(*line_buf);
				*line_buf = NULL;
				got_line = FALSE;
			}
		}
	}

	/* Check for and handle any messages from stdout */
	fp = win->cstdout;
	if(fp != NULL)
	{
		/* Check the prefix of each message to determine its
		 * type and handle it acclordingly
		 */
		gboolean got_line = FALSE;
#ifdef HAVE_LIBENDEAVOUR2
		const gint fd = (gint)fileno(fp);
#endif
		gchar	**line_buf = &win->cstdout_line_buf,
			*s,
			buf[32];
#ifdef HAVE_LIBENDEAVOUR2
		while(edv_poll_read(fd))
#else
		while(TRUE)
#endif
		{
			if(fgets((char *)buf, sizeof(buf), fp) == NULL)
				break;

			if(*line_buf != NULL)
				s = g_strconcat(
					*line_buf,
					buf,
					NULL
				);
			else
				s = g_strdup(buf);
			if(s != NULL)
			{
				g_free(*line_buf);
				*line_buf = s;
				s = (gchar *)strpbrk((char *)s, "\n\r");
				if(s != NULL)
				{
					*s = '\0';
					got_line = TRUE;
				}
			}
			if(got_line)
			{
				/* Seek past any spaces */
				const gchar *s = *line_buf;
				while(ISBLANK(*s))
					s++;

				/* Scanning file */
				if(g_strcasepfx(s, "Scanning:") ||
				   g_strcasepfx(s, "Scanning file:")
				)
				{
					/* Format is:
					 *
					 * <prefix> [type]: "<path>" <nscanned> <total>
					 */
					gchar	*msg,
						*shortened_path,
						*prefix = NULL,
						*type = NULL,
						*path = NULL;
					gulong	nscanned = 0l,
						total = 0l;
#ifdef HAVE_LIBENDEAVOUR2
					gchar	*arg;
#endif

#ifdef HAVE_LIBENDEAVOUR2
					/* Get the prefix */
					s = edv_strarg(
						s,
						&prefix,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(prefix != NULL)
					{
						gchar *s = strchr(prefix, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the type */
					s = edv_strarg(
						s,
						&type,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(type != NULL)
					{
						gchar *s = strchr(type, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the path */
					s = edv_strarg(
						s,
						&path,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);

					/* Get the number of files scanned */
					s = edv_strarg(
						s,
						&arg,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(arg != NULL)
					{
						scan_ctx->nscanned = nscanned = (gulong)ATOL(arg);
						g_free(arg);
					}

					/* Get the total files */
					s = edv_strarg(
						s,
						&arg,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(arg != NULL)
					{
						scan_ctx->total = total = (gulong)ATOL(arg);
						g_free(arg);
					}
#endif	/* HAVE_LIBENDEAVOUR2 */

					avs_win_update_title(win);

					/* Update the results status */
#ifdef HAVE_LIBENDEAVOUR2
					shortened_path = edv_path_shorten(
						path,
						25
					);
#else
					shortened_path = STRDUP(path);
#endif
					if(prefix != NULL)
						msg = g_strdup_printf(
							"%s %s (%ld/%ld): %s",
							prefix,
							type,
							nscanned + 1l,
							total,
							shortened_path
						);
					else
						msg = STRDUP(s);
					g_free(shortened_path);
					avs_win_set_results_status(win, msg);
					g_free(msg);

					/* Update the status bar message */
#ifdef HAVE_LIBENDEAVOUR2
					shortened_path = edv_path_shorten(
						path,
						40
					);
#else
					shortened_path = STRDUP(path);
#endif
					if(prefix != NULL)
						msg = g_strdup_printf(
							"%s %s (%ld/%ld): %s",
							prefix,
							type,
							nscanned + 1l,
							total,
							shortened_path
						);
					else
						msg = STRDUP(s);
					g_free(shortened_path);
					avs_win_set_status_message(win, msg, FALSE);
					g_free(msg);

					/* Update the status bar progress */
					avs_win_set_status_progress(
						win,
						(scan_ctx->total > 0l) ?
							((gfloat)scan_ctx->nscanned / (gfloat)scan_ctx->total) : 0.0f,
						TRUE
					);

					g_free(prefix);
					g_free(type);
					g_free(path);
				}
				/* Scanned */
				else if(g_strcasepfx(s, "Scanned:"))
				{
					/* Format is:
					 *
					 * <prefix>: <nscanned> <blocks_scanned> <total_blocks_scanned>
					 */
					gchar *prefix = NULL;
#ifdef HAVE_LIBENDEAVOUR2
					gchar *arg;
#endif

#ifdef HAVE_LIBENDEAVOUR2
					/* Get the prefix */
					s = edv_strarg(
						s,
						&prefix,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(prefix != NULL)
					{
						gchar *s = strchr(prefix, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the number of objects scanned */
					s = edv_strarg(
						s,
						&arg,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(arg != NULL)
					{
						scan_ctx->nscanned = (gulong)ATOL(arg);
						g_free(arg);
					}

					/* Get the data scanned for this object in blocks */
					s = edv_strarg(
						s,
						&arg,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(arg != NULL)
					{
						/* Ignored */
						g_free(arg);
					}

					/* Get the total data scanned in blocks */
					s = edv_strarg(
						s,
						&arg,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(arg != NULL)
					{
						scan_ctx->blocks_scanned = (gulong)ATOL(arg);
						g_free(arg);
					}
#endif	/* HAVE_LIBENDEAVOUR2 */

					avs_win_update_title(win);

					/* Update the status bar progress */
					avs_win_set_status_progress(
						win,
						(scan_ctx->total > 0l) ?
							((gfloat)scan_ctx->nscanned / (gfloat)scan_ctx->total) : 0.0f,
						TRUE
					);

					g_free(prefix);
				}
				/* Scanning directory? */
				else if(g_strcasepfx(s, "Scanning directory:"))
				{
#if 0
/* We ignore the scanning directory because it usually appears very
 * briefly and makes subsequent messages difficult to read in the cases
 * of scanning multiple subdirectories
 */
					/* Format is:
					 *
					 * <prefix>: "<path>"
					 */
					gchar	*shortened_path,
						*msg,
						*prefix = NULL,
						*type = NULL,
						*path = NULL;
#ifdef HAVE_LIBENDEAVOUR2
					/* Get the prefix */
					s = edv_strarg(
						s,
						&prefix,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(prefix != NULL)
					{
						gchar *s = strchr(prefix, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the type */
					s = edv_strarg(
						s,
						&type,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(type != NULL)
					{
						gchar *s = strchr(type, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the path */
					s = edv_strarg(
						s,
						&path,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
#endif	/* HAVE_LIBENDEAVOUR2 */

					avs_win_update_title(win);

					/* Update the results status */
#ifdef HAVE_LIBENDEAVOUR2
					shortened_path = edv_path_shorten(
						path,
						25
					);
#else
					shortened_path = STRDUP(path);
#endif
					if(prefix != NULL)
						msg = g_strdup_printf(
							"%s %s: %s",
							prefix,
							type,
							shortened_path
						);
					else
						msg = STRDUP(s);
					g_free(shortened_path);
					avs_win_set_results_status(win, msg);
					g_free(msg);

					/* Update the status bar message */
#ifdef HAVE_LIBENDEAVOUR2
					shortened_path = edv_path_shorten(
						path,
						40
					);
#else
					shortened_path = STRDUP(path);
#endif
					if(prefix != NULL)
						msg = g_strdup_printf(
							"%s %s: %s",
							prefix,
							type,
							shortened_path
						);
					else
						msg = g_strdup(s);
					g_free(shortened_path);
					avs_win_set_status_message(win, msg, TRUE);
					g_free(msg);

					g_free(prefix);
					g_free(type);
					g_free(path);
#endif
				}
				/* Clean */
				else if(g_strcasepfx(s, "Clean:"))
				{
					/* Format is:
					 *
					 * <prefix>: "<path>"
					 */
					gchar	*prefix = NULL,
						*path = NULL;
#ifdef HAVE_LIBENDEAVOUR2
					/* Get the prefix */
					s = edv_strarg(
						s,
						&prefix,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(prefix != NULL)
					{
						gchar *s = strchr(prefix, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the path */
					s = edv_strarg(
						s,
						&path,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
#endif	/* HAVE_LIBENDEAVOUR2 */

					/* Add this to the results list */
					if(path != NULL)
					{
						const gint new_row = avs_win_results_list_append(
							win,
							path,
							"Clean",
							FALSE
						);
						if(new_row > -1)
						{
							gtk_clist_moveto(
								clist,
								new_row, 0,
								1.0f, 0.0f
							);
						}
					}
					avs_win_update_title(win);
					g_free(prefix);
					g_free(path);
				}
				/* Infected */
				else if(g_strcasepfx(s, "Infected:"))
				{
					/* Format is:
					 *
					 * <prefix>: "<path>" <virus>
					 */
					gchar	*prefix = NULL,
						*path = NULL,
						*virus = NULL;
#ifdef HAVE_LIBENDEAVOUR2
					/* Get the prefix */
					s = edv_strarg(
						s,
						&prefix,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(prefix != NULL)
					{
						gchar *s = strchr(prefix, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the path */
					s = edv_strarg(
						s,
						&path,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);

					/* Get the virus name */
					virus = STRDUP(s);
#endif	/* HAVE_LIBENDEAVOUR2 */
					/* Count this as an infected file */
					scan_ctx->ninfected++;

					/* Add this to the results list */
					if(path != NULL)
					{
						const gint new_row = avs_win_results_list_append(
							win,
							path,
							virus,
							TRUE
						);
						if(new_row > -1)
						{
							gtk_clist_moveto(
								clist,
								new_row, 0,
								1.0f, 0.0f
							);
						}
					}
					avs_win_update_title(win);
					g_free(prefix);
					g_free(path);
					g_free(virus);
				}
				/* Problem */
				else if(g_strcasepfx(s, "Problem:"))
				{
					/* Format is:
					 *
					 * <prefix>: "<path>" <problem>
					 */
					gchar	*prefix = NULL,
						*path = NULL,
						*problem = NULL;
#ifdef HAVE_LIBENDEAVOUR2
					/* Get the prefpxi */
					s = edv_strarg(
						s,
						&prefix,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);
					if(prefix != NULL)
					{
						gchar *s = strchr(prefix, ':');
						if(s != NULL)
							*s = '\0';
					}

					/* Get the path */
					s = edv_strarg(
						s,
						&path,
						FALSE,	/* Maintain escapes */
						TRUE	/* Parse quotes */
					);

					/* Get the problem */
					problem = STRDUP(s);
#endif	/* HAVE_LIBENDEAVOUR2 */

					scan_ctx->nproblems++;

					/* Add this to the results list */
					if(path != NULL)
					{
						const gint new_row = avs_win_results_list_append(
							win,
							path,
							problem,
							FALSE
						);
						if(new_row > -1)
						{
							gtk_clist_moveto(
								clist,
								new_row, 0,
								1.0f, 0.0f
							);
						}
					}
					avs_win_update_title(win);
					g_free(prefix);
					g_free(path);
					g_free(problem);
				}
				/* All else print to the status bar's message */
				else
				{
					/* Print this line to the status bar's message */
					avs_win_set_status_message(
						win,
						s,
						TRUE
					);
				}
				g_free(*line_buf);
				*line_buf = NULL;
				got_line = FALSE;
			}
		}
	}

	/* Process has exited? */
	if(!avs_win_scan_process_is_running(win))
	{
		const gulong duration_t = scan_ctx->duration;
		gchar *duration_s;

		/* Mark that this timeout is no longer being called and
		 * decrement the passive busy count
		 */
		scan_ctx->toid = 0;
		avs_win_set_passive_busy(win, FALSE);

		/* Close the streams and delete any output files */
		avs_win_stop_scan_process(win);

		/* Format the duration string
		 *
		 * Longer than 1 hour?
		 */
		if(duration_t >= (60l * 60l))
			duration_s = g_strdup_printf(
				"%i:%.2i:%.2i",
				(gint)(duration_t / (60l * 60l)),
				(gint)((duration_t / 60) % 60l),
				(gint)(duration_t % 60l)
			);
		else	/* Less than 1 hour */
			duration_s = g_strdup_printf(
				"%i:%.2i",
				(gint)(duration_t / 60l),
				(gint)(duration_t % 60l)
			);

		/* Update the results status */
		if((scan_ctx->ninfected > 0l) || (scan_ctx->nproblems > 0l))
		{
			gchar *msg = g_strdup_printf(
				"%ld %s Found   %ld %s Found   Duration: %s",
				scan_ctx->ninfected,
				(scan_ctx->ninfected == 1l) ? "Virus" : "Viruses",
				scan_ctx->nproblems,
				(scan_ctx->nproblems == 1l) ? "Problem" : "Problems",
				duration_s
			);
			avs_win_set_results_status(win, msg);
			g_free(msg);
		}
		else
		{
			gchar *msg = g_strdup_printf(
				"Clean,    Duration: %s",
				duration_s
			);
			avs_win_set_results_status(win, msg);
			g_free(msg);
		}

		g_free(duration_s);

		if(scan_ctx->ninfected > 0l)
		{
#ifdef HAVE_LIBENDEAVOUR2
			/* Play the "completed" sound */
			edv_play_sound_completed(core->edv_ctx);
#endif
		}
		else
		{
#ifdef HAVE_LIBENDEAVOUR2
			/* Play "warning" sound */
			edv_play_sound_warning(core->edv_ctx);
#endif
		}

		/* Update the status message */
		avs_win_set_status_message(
			win,
"Scan done",
			FALSE
		);

		/* Raise the Win */
		avs_win_map(win);

		/* Display the report */
		WinResultsReportCB(win->toplevel, win);

		/* Do not call this function again */
		return(FALSE);
	}

	/* Update the progress only if the total number of files has
	 * not been received yet
	 */
	if((scan_ctx->total == 0l) && !win->paused)
		avs_win_set_status_progress(win, -1.0f, FALSE);

	/* Update the duration */
	if(scan_ctx->start_time > 0l)
	{
#ifdef HAVE_LIBENDEAVOUR2
		const gulong t = edv_time();
#else
		const gulong t = (gulong)time(NULL);
#endif
		scan_ctx->duration = (t > scan_ctx->start_time) ?
			(t - scan_ctx->start_time) : 0l;
		avs_win_set_status_time(
			win,
			scan_ctx->duration,
			FALSE
		);
	}

	return(TRUE);
}


/*
 *	Starts the virus database update process.
 *
 *	Any current virus database update process will be stopped first.
 */
void avs_win_start_db_update_net(AVSWin *win)
{
#if defined(HAVE_CLAMAV)
	struct stat stat_buf;
	const gint effective_uid = (gint)geteuid();
#endif
	gint pid;
	gchar		*prog_path,
			*cmd;
#ifdef HAVE_LIBENDEAVOUR2
	const gulong start_time = edv_time();
	EDVContext *edv_ctx;
#else
	const gulong start_time = (gulong)time(NULL);
#endif
	GtkWidget *toplevel;
	CfgList *cfg_list;
	AVSCore *core;

	if(win == NULL)
		return;

	toplevel = win->toplevel;
	core = AVS_CORE(win->core);
	cfg_list = core->cfg_list;
#ifdef HAVE_LIBENDEAVOUR2
	edv_ctx = core->edv_ctx;
#endif

	/* Clear the status message */
	avs_win_set_status_message(win, NULL, FALSE);

	/* Stop any existing virus database update process, close any
	 * streams to it, and remove any output files
	 */
	avs_win_stop_db_update_net(win);

	/* Get the path to the virus database update program and check
	 * if it is valid and set up properly
	 */
#if defined(HAVE_CLAMAV)
	/* Find the full path to freshclam */
# ifdef HAVE_LIBENDEAVOUR2
	prog_path = edv_which(NAME_FRESHCLAM);
	if(prog_path == NULL)
	{
		gchar *msg = g_strdup_printf(
"Unable to find the virus database update program:\n\
\n\
    %s",
			NAME_FRESHCLAM
		);
		edv_play_sound_warning(edv_ctx);
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Virus Database Update Failed",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		free(msg);
		CDialogSetTransientFor(NULL);
		return;
	}
# else
	prog_path = g_strconcat(
		"/usr/bin",
		G_DIR_SEPARATOR_S,
		NAME_FRESHCLAM,
		NULL
	);
# endif	/* HAVE_LIBENDEAVOUR2 */

	/* Check if freshclam exists */
	if(stat((const char *)prog_path, &stat_buf))
	{
		const gint error_code = (gint)errno;
		gchar *msg = g_strdup_printf(
"Unable to find the virus database update program:\n\
\n\
    %s\n\
\n\
%s.",
			prog_path,
			g_strerror(error_code)
		);
# ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_warning(edv_ctx);
# endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Virus Database Update Failed",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		free(msg);
		CDialogSetTransientFor(NULL);
		g_free(prog_path);
		return;
	}

	/* Check if freshclam's permissions are set up properly */
	if(!(stat_buf.st_mode & S_ISUID) &&
	   (effective_uid != (gint)stat_buf.st_uid) &&
	   (effective_uid != 0)
	)
	{
		gint response;
		gchar *msg = g_strdup_printf(
"You may not be able to update the virus database because\n\
the virus database update program:\n\
\n\
    %s\n\
\n\
Is not set SETUID and you are not the owner of it.\n\
\n\
Do you want to try and update the virus database anyway?",
			prog_path
		);
# ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_warning(edv_ctx);
# endif
		CDialogSetTransientFor(toplevel);
		response = CDialogGetResponse(
			"Virus Database Update Warning",
			msg,
			NULL,
			CDIALOG_ICON_WARNING,
			CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
			CDIALOG_BTNFLAG_NO
		);
		free(msg);
		CDialogSetTransientFor(NULL);
		if((response != CDIALOG_RESPONSE_YES) &&
		   (response != CDIALOG_RESPONSE_YES_TO_ALL)
		)
		{
			g_free(prog_path);
			return;
		}
	}
#else
# warning "avs_win_start_db_update_net() No compile-time antivirus database update method configured."
	prog_path = NULL;
#endif	/* HAVE_CLAMAV */

	/* Set the virus database update start time */
	win->db_update_net_start_time = start_time;

	/* Set the status message */
	avs_win_set_status_message(
		win,
		"Virus database update in progress...",
		FALSE
	);

	/* Set the virus database update monitor timeout callback and
	 * increment the passively busy count based on the setting of
	 * the timeout callback
	 */
	win->db_update_net_toid = gtk_timeout_add(
		250l,
		avs_win_db_update_monitor_timeout_cb, win
	);
	if(win->db_update_net_toid != 0)
		avs_win_set_passive_busy(win, TRUE);

	/* Format the virus database update command */
#if defined(HAVE_CLAMAV)
	cmd = g_strdup_printf(
		"\"%s\" --stdout --verbose",
		prog_path
	);
#else
	cmd = NULL;
#endif	/* HAVE_CLAMAV */

	/* Execute the virus database update command, open the streams
	 * to the new process and store them on the AVSWin
	 */
	win->db_update_net_pid = pid = avs_win_process_execute(
		win,
		cmd
	);
	if(pid <= 0) 
	{
		const gint error_code = (gint)errno;
		char *msg = g_strdup_printf(
"Unable to execute:\n\
\n\
    %s\n\
\n\
%s.",
			cmd,
			g_strerror(error_code)
		);
#ifdef HAVE_LIBENDEAVOUR2
		edv_play_sound_error(edv_ctx);
#endif
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
			"Virus Database Update Failed",
			msg,
			NULL,
			CDIALOG_ICON_ERROR,
			CDIALOG_BTNFLAG_OK,
			CDIALOG_BTNFLAG_OK
		);
		free(msg);
		CDialogSetTransientFor(NULL);

		avs_win_stop_db_update_net(win);
		avs_win_update_display(win);

		g_free(cmd);
		g_free(prog_path);
		return;
	}

	/* Delete the command */
	g_free(cmd);

	/* Reset the progress and time */
	avs_win_set_status_progress(win, -1.0f, FALSE);
	avs_win_set_status_time(win, 0l, FALSE);

	avs_win_update_title(win);
	avs_win_update_display(win);

	g_free(prog_path);
}

/*
 *	Stops the virus database update process and cleans up.
 *
 *	If the scan monitoring timeout was set then it will be removed
 *	and decrements a passive busy count.
 *
 *	Interruptes the virus database update process, closes any
 *	streams and deletes any output files.
 *
 *	Resets the status bar and scan values.
 */
void avs_win_stop_db_update_net(AVSWin *win)
{
	gboolean was_monitoring_process;

	if(win == NULL)
		return;

	/* Remove the virus database update timeout callback */
	if(win->db_update_net_toid != 0)
	{
		gtk_timeout_remove(win->db_update_net_toid);
		win->db_update_net_toid = 0;
		was_monitoring_process = TRUE;
	}
	else
		was_monitoring_process = FALSE;

	/* Interrupt the virus database update process */
	if(win->db_update_net_pid > 0)
	{
#ifdef SIGCONT
		if(win->paused)
			(void)kill(
				(int)win->db_update_net_pid,
				SIGCONT
			);
#endif
#ifdef SIGINT
		(void)kill(
			(int)win->db_update_net_pid,
			SIGINT
		);
#else
# warning "SIGINT not defined, virus database process stopping will not function."
#endif
		win->db_update_net_pid = 0;
	}

	/* Close the streams to the process and remove any output files */
	avs_win_delete_process_output(win);

	/* Reset the stop counter and paused state */
	win->stop_count = 0;
	win->paused = FALSE;
	win->db_update_net_start_time = 0l;

	/* Reset the progress and time */
	avs_win_set_status_progress(win, 0.0f, FALSE);
	avs_win_set_status_time(win, 0l, FALSE);

	avs_win_update_title(win);
	avs_win_update_display(win);

	/* Mark the AVSWin as no longer passively busy */
	if(was_monitoring_process)
		avs_win_set_passive_busy(win, FALSE);
}

/*
 *	Checks if the virus database update process is running.
 */
gboolean avs_win_db_update_net_process_is_running(AVSWin *win)
{
	if(win == NULL)
		return(FALSE);

	if(win->db_update_net_pid <= 0)
		return(FALSE);

#ifdef HAVE_LIBENDEAVOUR2
	return(edv_pid_exists(win->db_update_net_pid));
#else
	return(ExecProcessExists(win->db_update_net_pid));
#endif
}

/*
 *	Virus database update monitoring timeout callback.
 *
 *	Monitors the virus database update process and handles/displays
 *	any messages from it.
 */
static gint avs_win_db_update_monitor_timeout_cb(gpointer data)
{
	gint pid;
	FILE *fp;
	GtkWidget *toplevel;
	AVSCore *core;
	AVSWin *win = AVS_WIN(data);
	if(win == NULL)
		return(FALSE);

	pid = win->db_update_net_pid;
	toplevel = win->toplevel;
	core = AVS_CORE(win->core);

	/* Need to stop? */
	if(win->stop_count > 0)
	{
		/* Stop the virus database update
		 *
		 * Mark that this timeout is no longer being called and
		 * decrement the passive busy count
		 */
		win->db_update_net_toid = 0;
		avs_win_set_passive_busy(win, FALSE);

		avs_win_stop_db_update_net(win);

		avs_win_set_status_message(
			win,
			"Virus database update interrupted",
			FALSE
		);

		/* Do not call this function again */
		return(FALSE);
	}

	/* Check for any handle any messages from stdout */
	fp = win->cstdout;
	if(fp != NULL)
	{
		gboolean got_line = FALSE;
#ifdef HAVE_LIBENDEAVOUR2
		const gint fd = (gint)fileno(fp);
#endif
		gint c;
		gchar	**line_buf = &win->cstdout_line_buf,
			*s,
			buf[32],
			*buf_ptr, *buf_end;
#ifdef HAVE_LIBENDEAVOUR2
		while(edv_poll_read(fd))
#else
		while(TRUE)
#endif
		{
			/* Read using fgetc() instead of fgets() so that
			 * we can catch the carrage return ('\n') to
			 * properly display freshclam's progress
			 */
			buf_ptr = buf;
			buf_end = buf + sizeof(buf) - 1;
			do {
				c = (gint)fgetc(fp);
				if((c == EOF) || (c == '\0'))
					break;

				*buf_ptr = (gchar)c;
				buf_ptr++;

			} while((buf_ptr < buf_end) &&
				(c != '\r') && (c != '\n') &&
#ifdef HAVE_LIBENDEAVOUR2
				edv_poll_read(fd)
#else
				TRUE
#endif
			);

			*buf_ptr = '\0';

			if(*line_buf != NULL)
				s = g_strconcat(
					*line_buf,
					buf,
					NULL
				);
			else
				s = g_strdup(buf);
			if(s != NULL)
			{
				g_free(*line_buf);
				*line_buf = s;
				s = (gchar *)strpbrk((char *)s, "\n\r");
				if(s != NULL)
				{
					*s = '\0';
					got_line = TRUE;
				}
			}
			if(got_line)
			{
				const gchar *s2;

				/* Seek past any spaces */
				const gchar *s = *line_buf;
				while(ISBLANK(*s))
					s++;

				/* Look for the progress in this string
				 * by searching for the '%' character
				 */
				s2 = (gchar *)strchr((char *)s, '%');
				if(s2 != NULL)
				{
					/* Seek to the start of the
					 * percent value
					 */
					if(s2 > s)
						s2--;
					while(s2 > s)
					{
						if(!isdigit((int)*(s2 - 1)))
							break;
						s2--;
					}

					/* Update the progress value on
					 * the status bar and then the
					 * title (since setting the title
					 * takes the already set progress
					 * value on the status bar)
					 */
					if(!win->paused)
					{
						avs_win_set_status_progress(
							win,
							(gfloat)ATOI(s2) / 100.0f,
							FALSE
						);
						avs_win_update_title(win);
					}
				}

				avs_win_set_status_message(
					win,
					s,
					TRUE
				);

				g_free(*line_buf);
				*line_buf = NULL;
				got_line = FALSE;
			}

			if(c == EOF)
				break;
		}
	}

	/* Check for any handle any messages from stderr */
	fp = win->cstderr;
	if(fp != NULL)
	{
		gboolean got_line = FALSE;
#ifdef HAVE_LIBENDEAVOUR2
		const gint fd = (gint)fileno(fp);
#endif
		gchar	**line_buf = &win->cstderr_line_buf,
			*s,
			buf[32];
#ifdef HAVE_LIBENDEAVOUR2
		while(edv_poll_read(fd))
#else
		while(TRUE)
#endif
		{
			if(fgets((char *)buf, sizeof(buf), fp) == NULL)
				break;

			if(*line_buf != NULL)
				s = g_strconcat(
					*line_buf,
					buf,
					NULL
				);
			else
				s = g_strdup(buf);
			if(s != NULL)
			{
				g_free(*line_buf);
				*line_buf = s;
				s = (gchar *)strpbrk((char *)s, "\n\r");
				if(s != NULL)
				{
					*s = '\0';
					got_line = TRUE;
				}
			}
			if(got_line)
			{
				gint response;
				gchar *msg;

				/* Seek past any spaces */
				const gchar *s = *line_buf;
				while(ISBLANK(*s))
					s++;

				/* Print the error message to the user */
#ifdef HAVE_LIBENDEAVOUR2
				edv_play_sound_warning(core->edv_ctx);
#endif
				CDialogSetTransientFor(toplevel);
				msg = g_strdup_printf(
"%s\n\
\n\
Do you want to stop the virus database update?",
					s
				);
				response = CDialogGetResponse(
					"Update Virus Database Error",
					msg,
					NULL,
					CDIALOG_ICON_WARNING,
					CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
					CDIALOG_BTNFLAG_NO
				);
				g_free(msg);
				CDialogSetTransientFor(NULL);
				if(response == CDIALOG_RESPONSE_YES)
				{
			                /* Stop the virus database update */
					win->db_update_net_toid = 0;
			                avs_win_set_passive_busy(win, FALSE);
					avs_win_stop_db_update_net(win);
					avs_win_set_status_message(
						win,
"Virus database update interrupted",
						FALSE
					);
					g_free(*line_buf);
					*line_buf = NULL;

					/* Do not call this function again */
					return(FALSE);
				}

				g_free(*line_buf);
				*line_buf = NULL;
				got_line = FALSE;
			}
		}
	}


	/* Has the virus database update process exited? */ 
	if(!avs_win_db_update_net_process_is_running(win))
	{
		/* Record the virus database update time */
		CFGItemListSetValueL(
			core->cfg_list, CFG_PARM_DB_LAST_UPDATE,
#ifdef HAVE_LIBENDEAVOUR2
			edv_time(),
#else
			(gulong)time(NULL),
#endif
			FALSE
		);

#ifdef HAVE_LIBENDEAVOUR2
		/* Play the "completed" sound */
		edv_play_sound_completed(core->edv_ctx);
#endif

		/* Stop the virus database update
		 *
		 * Mark that this timeout is no longer being called and
		 * decrement the passive busy count
		 */
		win->db_update_net_toid = 0;
		avs_win_set_passive_busy(win, FALSE);

		/* Close the streams and delete any output files */
		avs_win_stop_db_update_net(win);

		avs_win_set_status_message(
			win,
			"Virus database update done",
			TRUE
		);

		/* Queue update the displayed values for the Virus Database */
		avs_win_db_queue_update(win);

		/* Do not call this function again */
		return(FALSE);
	}

	/* Update the progress only if no known progress value has
	 * been reported yet
	 */
	if((win->status_bar_progress_last < 0.0f) && !win->paused)
		avs_win_set_status_progress(win, -1.0f, FALSE);

	/* Update the time */
	if(win->db_update_net_start_time > 0l)
	{
#ifdef HAVE_LIBENDEAVOUR2
		const gulong t = edv_time();
#else
		const gulong t = (gulong)time(NULL);
#endif
		avs_win_set_status_time(
			win,
			(t > win->db_update_net_start_time) ?
				(t - win->db_update_net_start_time) : 0l,
			FALSE
		);
	}

	return(TRUE);
}
